import csv
import json
import logging
from dataclasses import dataclass
from datetime import datetime
from enum import auto
from time import sleep
from typing import TYPE_CHECKING

import humps
import pandas as pd

from mstrio import config
from mstrio.api import license as license_api
from mstrio.connection import Connection
from mstrio.server import Environment
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import (
    Dictable,
    IServerError,
    camel_to_snake,
    key_fn_for_sort_object_properties,
)
from mstrio.utils.version_helper import class_version_handler

if TYPE_CHECKING:
    from mstrio.access_and_security import Privilege

logger = logging.getLogger(__name__)
WS_TIME_FORMAT = '%Y-%m-%d %I:%M:%S %p'


class InstallationUse(AutoName):
    """Enumeration constant indicating installation use"""

    DEVELOPMENT = auto()
    TESTING = auto()
    PRODUCTION = auto()
    TRAINING = auto()
    OTHER = auto()


@dataclass
class ContactInformation(Dictable):
    """Contact information for license management. This class represents
    contact information required for license operations.
    """

    department: str
    first_name: str
    last_name: str
    job_title: str
    phone: str
    email: str
    confirm_email: str
    address: str
    city: str
    postal: str
    country: str
    company: str
    use_personal_info: bool = False
    state: str | None = None

    def __post_init__(self):
        if self.email != self.confirm_email:
            raise ValueError("Email address and confirmation email do not match")


@dataclass
class ActivationInfo(Dictable):
    """Class that represents license activation information. It contains details
    about the activation status, activation and deactivation codes, and other
    related information.
    """

    activated: bool
    activation_count: int
    activation_needed: bool
    user_provided_info: dict | None = None
    activation_code: str | None = None
    deactivation_code: str | None = None
    activation_date: str | None = None


@dataclass
class MachineInfo(Dictable):
    """
    Class representing system hardware and operating system information
    for a machine.

    """

    cpu_speed: str
    cpu_type: str
    OS: str
    host_id: str
    kernel_release: str
    machine_name: str
    num_physical_cpus: int
    system_language: str


@dataclass
class PrivilegeInfo(Dictable):
    """Class representing a privilege information in the license management
    system.
    """

    privilege: "Privilege"
    sources: list[str] | None = None


@dataclass
class UserLicense(Dictable):
    """Class representing a user license in the license management system."""

    id: str
    name: str
    enabled: bool
    date_created: str
    date_modified: str
    ldap: bool | None = None
    privileges: list[PrivilegeInfo] | None = None


@dataclass
class Product(Dictable):
    """Class encapsulating details about a licensed product within the Strategy
    ecosystem.
    """

    count: int
    duration: int
    id: int
    name: str
    period: int
    type: str
    module: str
    _connection: Connection
    in_compliance: bool | None = None
    license_in_use: int | None = None
    users: list[UserLicense] | None = None

    PRODUCT_NAME_MAP = {
        0: "RESERVED",
        6: "SERVER_REPORTER",
        17: "GATEWAY_EMM_AIRWATCH_BY_VMWARE",
        23: "CLIENT_APPLICATION_OFFICE",
        30: "CLIENT_HYPER_API",
        32: "PLACEHOLDER_1",
        33: "PLACEHOLDER_2",
        34: "CLIENT_APPLICATION_RSTUDIO",
        35: "PRODUCT_01",
        36: "PRODUCT_02",
        37: "PRODUCT_03",
        38: "PRODUCT_04",
        39: "PRODUCT_05",
        40: "PRODUCT_06",
    }

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"name='{self.name}', "
            f"type='{self.type}', "
            f"license_in_use='{self.license_in_use}', "
            f"Entitlements count='{self.count}', "
            f"in_compliance='{self.in_compliance}')"
        )

    def fetch_users_privileges(self) -> None:
        """Fetch users privileges for the product."""
        from mstrio.access_and_security import Privilege

        if self.users:
            for user in self.users:
                user_privileges = camel_to_snake(
                    license_api.get_privileges_for_user(
                        connection=self._connection,
                        id=user.id,
                        license_product=self._convert_product_name_to_constant_format(),
                    ).json()
                ).get('pws')
                for record in user_privileges:
                    priv_obj = Privilege.from_dict(
                        record.get('privilege'), self._connection
                    )

                    privilege_info = PrivilegeInfo(priv_obj, record.get('sources'))
                    user.privileges.append(privilege_info)

    def _convert_product_name_to_constant_format(self) -> str:
        if self.id in self.PRODUCT_NAME_MAP:
            return self.PRODUCT_NAME_MAP[self.id]
        first_step = self.name.replace("-", "").replace(" ", "")
        result = humps.decamelize(first_step)
        return result.upper()


@dataclass
class Record(Dictable):
    """Class representing a record in the license management history"""

    activity: str
    build: str
    code: str
    contract_id: str
    issue_date: str
    key: str
    key_group: str
    products_install: list
    products_repair: list
    products_uninstall: list
    products_update: list
    release: str
    source: str
    state: str
    time: str

    def __repr__(self) -> str:
        time = datetime.strptime(self.time, '%Y-%m-%dT%H:%M:%SZ')
        formatted_time = time.strftime(WS_TIME_FORMAT)
        return (
            f"{self.__class__.__name__}("
            f"time='{formatted_time}', "
            f"activity='{self.activity}', "
            f"source='{self.source}', "
            f"state='{self.state}')"
        )


@class_version_handler('11.3.1200')
class License(Dictable):
    """License class for Strategy license management.

    This class provides functionality to manage Strategy licenses, including:
    - Retrieving license information
    - Registering new license keys
    - Activating and deactivating licenses
    - Managing activation requests
    - Generating and downloading activation XML files

    Attributes:
        bypass_audit (bool): Flag indicating if audit functionality is bypassed.
        bypass_cpu_controls (bool): Flag indicating if CPU controls are
            bypassed.
        contract_id (str): The contract ID of the license.
        disable_activation (bool): Flag indicating if activation is disabled.
        enable_daily_report (bool): Flag indicating if daily reporting is
            enabled.
        enable_key_level_evaluation (bool): Flag for key level evaluation.
        installation_version (str): Version of the installation.
        internal (bool): Flag indicating if the license is internal.
        issue_date (str): Date when the license was issued.
        license_key_group (str): License key group information.
        machine_info (MachineInfo): Information about the machine where license
            is installed.
        platform (str): Platform information for the license.
        pre_release (bool): Flag indicating if it's a pre-release version.
        products (list): List of products included in the license.
        register_date (str): Date when the license was registered.
        registration_disabled (bool): Flag indicating if registration is
            disabled.
        status (str): Current status of the license.
        edition (str): Edition of the license.
        cpu_allowance (int): CPU allowance for the license.
        AWS (bool): Flag indicating if it's an AWS installation.
        act_aware_products (list): List of activation aware products.
        DSI (bool): Flag indicating if DSI is enabled.
        installation_type (str): Type of installation.
        enabled_poc (bool): Flag indicating if PoC is enabled.
        key (str): License key.
        v_cpus (int): Number of virtual CPUs.
        activation_info (ActivationInfo): Information about license activation.
        compliance_time (str): Time of the last compliance check.
        compliance_type (str): Type of the last compliance check.
        in_compliance (bool): Flag indicating if the license is in compliance.
        compliance_running (bool): Flag indicating if compliance check is
            currently running.
        audit_time (str): Time of the last audit.
    """

    FIELDNAMES = [
        "Product",
        "User Full Name",
        "User ID",
        "User State",
        "Creation Date (User)",
        "Last Modified (User)",
        "LDAP (User)",
        "Privilege Name",
        "Privilege ID",
        "Inherited from User Group",
        "Inherited from Security Role",
        "Project",
    ]

    def __init__(self, connection: Connection, node_name: str | None = None):
        """Initialize License object.

        Args:
            connection: Strategy One connection object returned by
                `connection.Connection()`.
            node_name (str, optional): Name of the node to get license
                information from. On the one node environment can be None.
        """
        self._connection = connection
        self._node_name = None
        self._activation_info = None
        self._compliance_time = None
        self._compliance_type = None
        self._in_compliance = None
        self._compliance_running = None
        self._audit_time = None
        self._check_node_name(node_name)
        resp = license_api.get_license(
            connection=self._connection, node_name=self._node_name
        )
        retr = camel_to_snake(resp.json())
        self._init_variables(**retr)
        if config.verbose:
            logger.info(
                f"License object successfully initialized for node '{self._node_name}'."
            )

    def _init_variables(self, **kwargs) -> None:
        """This function initializes variables using keyword arguments
        passed to it.
        """
        if kwargs.get('machine_info'):
            physical_cpus = kwargs['machine_info'].get('num_physical_cp_us')
            kwargs['machine_info']['num_physical_cpus'] = physical_cpus
            del kwargs['machine_info']['num_physical_cp_us']

        self.bypass_audit = kwargs.get('bypass_audit')
        self.bypass_cpu_controls = kwargs.get('bypass_cpu_controls')
        self.contract_id = kwargs.get('contract_id')
        self.disable_activation = kwargs.get('disable_activation')
        self.enable_daily_report = kwargs.get('enable_daily_report')
        self.enable_key_level_evaluation = kwargs.get('enable_key_level_evaluation')
        self.installation_version = kwargs.get('installation_version')
        self.internal = kwargs.get('internal')
        self.issue_date = kwargs.get('issue_date')
        self.license_key_group = kwargs.get('license_key_group')
        self.machine_info = (
            MachineInfo(**kwargs['machine_info'])
            if kwargs.get('machine_info')
            else None
        )
        self.platform = kwargs.get('platform')
        self.pre_release = kwargs.get('pre_release')
        self.products = [
            Product(module=key, _connection=self._connection, **product_info)
            for key, products_list in kwargs.get('products', {}).items()
            for product_info in products_list
        ]
        self.register_date = kwargs.get('register_date')
        self.registration_disabled = kwargs.get('registration_disabled')
        self.status = kwargs.get('status')
        self.edition = kwargs.get('edition')
        self.cpu_allowance = kwargs.get('cpu_allowance')
        self.AWS = kwargs.get('AWS')
        self.act_aware_products = kwargs.get('act_aware_products')
        self.DSI = kwargs.get('DSI')
        self.installation_type = kwargs.get('installation_type')
        self.enabled_poc = kwargs.get('enabled_poc')
        self.key = kwargs.get('key')
        self.v_cpus = kwargs.get('v_cp_us')

    def register_new_key(self, key: str) -> None:
        """Register a new license key. After registering a new key, the server
        must be restarted to apply the changes.

        Args:
            key (str): License key to register.
        """
        operations_list = [
            {
                "op": "replace",
                "path": "/key",
                "value": key,
            }
        ]
        license_api.update_license_information(
            self._connection, self._node_name, operations_list
        )
        if config.verbose:
            logger.info(
                "License key registered successfully. Now please restart the "
                "I-Server to apply the changes."
            )
        self._fetch()

    def activate(self, activation_code: str) -> None:
        """Activate the license.

        Args:
            activation_code: Activation code. To obtain the activation code you
                need to upload xml activation file on the Strategy licensing
                site.
        """
        operations_list = [
            {
                "op": "add",
                "path": "/activationCode",
                "value": activation_code,
            },
            {"op": "replace", "path": "/activated", "value": True},
        ]
        license_api.update_license_activation(
            self._connection, self._node_name, operations_list
        )
        if config.verbose:
            logger.info("License activated successfully.")
        self._fetch()

    def deactivate(self) -> None:
        """Deactivate the license."""
        operations_list = [{"op": "replace", "path": "/activated", "value": False}]
        license_api.update_license_activation(
            self._connection, self._node_name, operations_list
        )
        if config.verbose:
            logger.info("License deactivated successfully.")
        self._fetch()

    def list_details(self) -> dict:
        """Return a dictionary with license details.

        Returns:
            Dictionary with license details excluding products information.
        """
        attributes = {
            key: self.__dict__[key]
            for key in self.__dict__
            if not key.startswith('_') and key != 'products'
        }

        return {
            key: attributes[key]
            for key in sorted(attributes, key=key_fn_for_sort_object_properties)
        }

    def list_products(
        self,
        module: str = 'intelligence_module',
        product_id: int | list[int] | None = None,
        name: str | None = None,
    ) -> list:
        """Return a list of licensed products.
        Args:
            module (str, optional): Module name to filter products list.
                Default is 'intelligence_module', input 'all' to get list of
                products in every module.
            product_id (int | list[int], optional): Product IDs to filter
                products list. If provided, it should be a single ID or a list
                of IDs.
            name (str, optional): Product name to filter products list.
        Returns:
            List of licensed products.

        """
        if product_id is not None and name:
            raise ValueError(
                "You can provide either 'id' or 'name' to filter products, not both."
            )
        products = None
        if module == 'all':
            products = self.products
        else:
            if not any(product.module == module for product in self.products):
                raise ValueError(
                    f"Module '{module}' not found. Available modules: "
                    f"{', '.join({product.module for product in self.products})}"
                )
            products = [
                product for product in self.products if product.module == module
            ]
        if product_id is not None:
            if isinstance(product_id, int):
                product_id = [product_id]
            products = [product for product in products if product.id in product_id]
        if name:
            products = [
                product for product in products if name.lower() in product.name.lower()
            ]
        return sorted(products, key=lambda x: x.name)

    def list_properties(self) -> dict:
        """Return a dictionary with all license properties.

        Returns:
            Dictionary with all license properties.
        """
        attributes = {
            key: self.__dict__[key] for key in self.__dict__ if not key.startswith('_')
        }
        attributes = {
            **attributes,
            'connection_url': self._connection.base_url,
            'node_name': self._node_name,
            'activation_info': self.activation_info,
        }

        return {
            key: attributes[key]
            for key in sorted(attributes, key=key_fn_for_sort_object_properties)
        }

    def fetch(self) -> None:
        """Fetch the current license information from the environment."""
        self._fetch()
        if config.verbose:
            logger.info("License information fetched successfully.")

    def update_activation_request_information(
        self,
        location: str,
        installation_use: InstallationUse,
        customer_contact: ContactInformation | dict | str,
        employee_of_licensed_company: bool = True,
        installer_contact: ContactInformation | dict | str | None = None,
    ) -> None:
        """Update activation request information for the license.

        Args:
            location (str): Physical location of the installation.
            installation_use (InstallationUse) : Purpose of the installation
                (development, testing, production, training, or other).
            customer_contact (ContactInformation | dict | str): Contact
                information of the customer. Can be a ContactInformation object,
                dictionary or path to a file (JSON or CSV).
            employee_of_licensed_company (bool): Whether the person performing
                installation is an employee of licensed company or not.
                If False, installer_contact must be provided.
            installer_contact (ContactInformation | dict | str, optional):
                Contact information of the installer. Can be a
                ContactInformation object, dictionary or path to a file (JSON
                or CSV). Required when employee_of_licensed_company is False.
        """
        operations_list = [
            {
                "op": "add",
                "path": "/userProvidedInfo",
                "value": {
                    "installationName": self._node_name,
                    "installationLocation": location,
                    "installationUse": installation_use.value,
                    "contact": {},
                },
            }
        ]
        self._prepare_body_for_information_update(
            operations_list, customer_contact, 'customer'
        )
        if not employee_of_licensed_company:
            if not installer_contact:
                raise ValueError(
                    "Installer contact information is required when you are not an "
                    "employee of licensed company."
                )
            self._prepare_body_for_information_update(
                operations_list, installer_contact, 'installer'
            )
        else:
            operations_list[0]['value']['contact']['customer'][
                'company'
            ] = 'MicroStrategy'
        license_api.update_license_activation(
            self._connection, self._node_name, operations_list
        )
        if config.verbose:
            logger.info(
                "Activation request information updated successfully. "
                "Now you can download the activation XML file."
            )

        self.get_activation_info()

    def get_activation_xml_file(self, save_path: str) -> None:
        """Get the activation XML file to upload to the Strategy licensing site.

        Args:
            save_path (str): Path where the XML file will be saved.
                Must end with '.xml'.

        Raises:
            ValueError: If save_path doesn't end with '.xml' or if there's an
                error saving the file.
        """
        resp = license_api.get_activation_xml_file(self._connection, self._node_name)
        if not save_path.lower().endswith('.xml'):
            raise ValueError(
                "Save path must end with '.xml'. Please provide a valid file path."
            )
        try:
            with open(save_path, "w", encoding="utf-8") as f:
                f.write(resp.json())
        except Exception as e:
            raise ValueError(f"Error saving activation XML file: {str(e)}")
        if config.verbose:
            logger.info(
                f"Activation XML file saved successfully to '{save_path}'. "
                "Please upload it to Strategy licensing site."
            )

    def get_activation_info(self) -> ActivationInfo:
        """Get information about the license activation.

        Returns:
            ActivationInfo object with license activation details.
        """
        resp = license_api.get_license_activation_info(
            connection=self._connection, node_name=self._node_name
        )
        self._activation_info = ActivationInfo.from_dict(resp.json())
        return self._activation_info

    def get_history(self) -> list[Record]:
        """Get the license history.

        Returns:
            List of Record objects representing the license history.
        """
        resp = license_api.get_license_history(
            connection=self._connection, node_name=self._node_name
        )
        history = resp.json().get('activities', [])
        return sorted(
            (Record.from_dict(record) for record in history),
            key=lambda x: x.time,
        )

    def run_compliance_check(self) -> None:
        """Run a compliance check on the license."""
        license_api.run_compliance_check(connection=self._connection)
        if config.verbose:
            logger.info("Compliance check request sent successfully.")

    def get_compliance_check(self) -> None:
        """Get the compliance check details."""
        resp = camel_to_snake(
            license_api.get_compliance_check(connection=self._connection).json()
        )
        self._compliance_time = resp.get('compliance_time')
        self._compliance_type = resp.get('compliance_type')
        self._in_compliance = resp.get('in_compliance')
        if self._compliance_time:
            self._compliance_time = datetime.strptime(
                self._compliance_time, '%Y-%m-%dT%H:%M:%S.%fZ'
            ).strftime(WS_TIME_FORMAT)
        for elem in resp.get('compliance_result', []):
            product = list(
                filter(
                    lambda x, current_elem=elem: x.id == current_elem.get('type'),
                    self.products,
                )
            )
            for found_product in product:
                found_product.in_compliance = elem.get('inCompliance')
                found_product.license_in_use = elem.get('currentUsage')
        if (
            self._compliance_type == 0
            and self._compliance_time is None
            and self._in_compliance is False
        ):
            logger.warning(
                'Compliance check is unavailable when license key contains '
                'mixed license types for several products.'
            )
        if config.verbose:
            logger.info("Compliance check results fetched successfully.")

    def get_compliance_check_status(self) -> None:
        """Get the compliance check status."""
        resp = license_api.get_compliance_check_status(connection=self._connection)
        self._compliance_running = resp.json().get('jobRunning')

    def run_and_get_compliance_check(
        self, timeout: int = 30, retry_interval: int = 1
    ) -> None:
        """Run a compliance check and get the results with a timeout.

        Args:
            timeout (int, optional): Maximum time (in seconds) to wait for
                compliance check to complete.
            retry_interval (int, optional): Time interval (in seconds) between
                retries.
        """
        initial_compliance_time = self.compliance_time
        self.run_compliance_check()
        start_time = datetime.now()

        # Wait for compliance check status to change to not running
        while (
            self.compliance_running
            and (datetime.now() - start_time).total_seconds() < timeout
        ):
            sleep(retry_interval)
            self.get_compliance_check_status()

        if self.compliance_running:
            raise TimeoutError(
                "Compliance check did not complete within the allowed timeout."
            )

        # Fetch results and wait for compliance_time to update
        while (
            self.compliance_time == initial_compliance_time
            and (datetime.now() - start_time).total_seconds() < timeout
        ):
            sleep(retry_interval)
            self.get_compliance_check()

        if self.compliance_time == initial_compliance_time:
            raise TimeoutError(
                "Compliance check results were not updated within the allowed timeout."
            )

    def run_audit(self) -> dict:
        """Run a license audit.
        Returns:
            dict: Dictionary about the audit request.
        """
        resp = license_api.run_audit(connection=self._connection)
        if config.verbose:
            logger.info("License audit request sent successfully.")
        return camel_to_snake(resp.json())

    def get_audit(self) -> None:
        """Get the audit results."""
        resp = camel_to_snake(license_api.get_audit(connection=self._connection).json())
        self._audit_time = resp.get('audit_time')
        if self._audit_time:
            self._audit_time = datetime.strptime(
                self._audit_time, '%Y-%m-%dT%H:%M:%S.%fZ'
            ).strftime(WS_TIME_FORMAT)

        product_map = {product.id: product for product in self.products}
        for product in self.products:
            product.users = []

        for elem in resp.get('user_licenses', []):
            user_license = UserLicense(
                id=elem.get('id'),
                name=elem.get('name'),
                enabled=elem.get('enabled'),
                date_created=elem.get('creationTime'),
                date_modified=elem.get('modificationTime'),
                ldap=elem.get('LDAPUser'),
                privileges=[],
            )
            for product_id in elem.get('licenseTypes', []):
                if product := product_map.get(product_id):
                    product.users.append(user_license)
        if config.verbose:
            logger.info("License audit results fetched successfully.")

    def run_and_get_audit(self, timeout: int = 30, retry_interval: int = 1) -> None:
        """Run a license audit and get the results with a timeout.

        Args:
            timeout (int, optional): Maximum time (in seconds) to wait for
                the audit to complete.
            retry_interval (int, optional): Time interval (in seconds) between
                retries.
        """
        initial_audit_time = self.audit_time
        self.run_audit()
        start_time = datetime.now()

        # Wait for audit time to update
        while (
            self.audit_time == initial_audit_time
            and (datetime.now() - start_time).total_seconds() < timeout
        ):
            sleep(retry_interval)
            self.get_audit()

        if self.audit_time == initial_audit_time:
            raise TimeoutError(
                "Audit results were not updated within the allowed timeout."
            )

    def fetch_all_products_users_privileges(self) -> None:
        """Fetch all products and their users' privileges."""
        if not self._audit_time:
            self.get_audit()
        for product in self.products:
            try:
                product.fetch_users_privileges()
            except IServerError as e:
                logger.error(
                    "Error fetching user privileges for product '%s': %s",
                    product.name,
                    e,
                )
        if config.verbose:
            logger.info(
                "All products and their users' privileges fetched successfully."
            )

    def get_new_key_entitlements(self, key: str) -> dict:
        """Get the entitlements for a new license key.

        Args:
            key (str): License key to get entitlements for.

        Returns:
            dict: Dictionary with the entitlements for the new license key.
        """
        resp = license_api.get_license_entitlements(
            connection=self._connection, node_name=self._node_name, license_key=key
        )
        return resp.json()

    def export_audit_to_csv(self, file_path: str) -> pd.DataFrame:
        """Export the audit results to a CSV file.

        IMPORTANT: To export detailed audit results, you need to fetch all
        products and their users' privileges first using the
        `fetch_all_products_users_privileges` method.

        Args:
            file_path (str): Path to the CSV file where the audit results will
                be saved.
        Returns:
            pd.DataFrame: DataFrame containing the audit results.
                In case of detailed audit results, if privilege is granted
                directly to the user, the columns 'Inherited from User Group',
                'Inherited from Security Role' and 'Project' will be set to
                '--'.
                If privilege is inherited from a UserGroup, the column
                'Inherited from User Group' will be set to the name of the
                UserGroup, columns 'Inherited from Security Role' and 'Project'
                will be set to '--'.
                If privilege is inherited from a SecurityRole, the column
                'Inherited from User Group' will be set to '--', columns
                'Inherited from Security Role' will be set to the name of the
                SecurityRole and 'Project' will be set to the name of the
                project the SecurityRole is assigned to.
        """
        if not file_path.lower().endswith('.csv'):
            raise ValueError("File path must end with '.csv'.")
        df = self.export_audit_to_dataframe(file_path)
        return df

    def export_audit_to_dataframe(self, file_path: str | None = None) -> pd.DataFrame:
        """Create and return a DataFrame containing the audit results.
        Optionally, save the DataFrame to a file.

        IMPORTANT: To export detailed audit results, you need to fetch all
        products and their users' privileges first using the
        `fetch_all_products_users_privileges` method.

        Args:
            file_path (str, optional): Path to the file where the DataFrame
                will be saved. If None, the DataFrame will not be saved. It
                can be a path to a .pkl, .json or .csv file.
        Returns:
            pd.DataFrame: DataFrame containing the audit results.
                In case of detailed audit results, if privilege is granted
                directly to the user, the columns 'Inherited from User Group',
                'Inherited from Security Role' and 'Project' will be set to
                '--'.
                If privilege is inherited from a UserGroup, the column
                'Inherited from User Group' will be set to the name of the
                UserGroup, columns 'Inherited from Security Role' and 'Project'
                will be set to '--'.
                If privilege is inherited from a SecurityRole, the column
                'Inherited from User Group' will be set to '--', columns
                'Inherited from Security Role' will be set to the name of the
                SecurityRole and 'Project' will be set to the name of the
                project the SecurityRole is assigned to.
        """

        rows = self._generate_audit_data()
        df = pd.DataFrame(rows, columns=self.FIELDNAMES)
        if file_path:
            if file_path.lower().endswith('.pkl'):
                df.to_pickle(file_path)
            elif file_path.lower().endswith('.json'):
                df.to_json(file_path, orient='records', lines=True)
            elif file_path.lower().endswith('.csv'):
                df.to_csv(file_path, index=False, encoding='utf-8')
            else:
                raise ValueError("File path must end with '.pkl', '.json' or '.csv'.")
            if config.verbose:
                logger.info(f"Audit results exported successfully to '{file_path}'.")
        return df

    def _generate_audit_data(self) -> list[dict]:
        """Generate audit data as a list of dictionaries."""
        if not self.audit_time:
            self.get_audit()

        rows = []
        for product in self.products:
            for user in product.users:
                if user.privileges:
                    rows.extend(self._generate_rows_for_user_privileges(product, user))
                else:
                    rows.append(self._build_na_record(product, user))
        return rows

    def _generate_rows_for_user_privileges(
        self, product: Product, user: UserLicense
    ) -> list[dict]:
        """Generate rows for user privileges."""
        rows = []
        for privilege_record in user.privileges:
            for source in privilege_record.sources:
                user_group, security_role, project = self._define_source(source)
                rows.append(
                    {
                        self.FIELDNAMES[0]: product.name,
                        self.FIELDNAMES[1]: user.name,
                        self.FIELDNAMES[2]: user.id,
                        self.FIELDNAMES[3]: "Enabled" if user.enabled else "Disabled",
                        self.FIELDNAMES[4]: user.date_created,
                        self.FIELDNAMES[5]: user.date_modified,
                        self.FIELDNAMES[6]: user.ldap,
                        self.FIELDNAMES[7]: privilege_record.privilege.name,
                        self.FIELDNAMES[8]: privilege_record.privilege.id,
                        self.FIELDNAMES[9]: user_group,
                        self.FIELDNAMES[10]: security_role,
                        self.FIELDNAMES[11]: project,
                    }
                )
        return rows

    def _build_na_record(self, product: Product, user: UserLicense) -> dict:
        """
        Build a record with N/A placeholders when privileges aren't present.
        """
        return {
            self.FIELDNAMES[0]: product.name,
            self.FIELDNAMES[1]: user.name,
            self.FIELDNAMES[2]: user.id,
            self.FIELDNAMES[3]: "Enabled" if user.enabled else "Disabled",
            self.FIELDNAMES[4]: user.date_created,
            self.FIELDNAMES[5]: user.date_modified,
            self.FIELDNAMES[6]: user.ldap,
            self.FIELDNAMES[7]: "N/A",
            self.FIELDNAMES[8]: "N/A",
            self.FIELDNAMES[9]: "N/A",
            self.FIELDNAMES[10]: "N/A",
            self.FIELDNAMES[11]: "N/A",
        }

    def _define_source(self, source: dict) -> tuple[str, str, str]:
        """Define the source of the license.

        Args:
            source (dict): Dictionary containing source information.

        Returns:
            tuple[str, str, str]: A tuple containing the UserGroup, SecurityRole
                and Project information.
        """
        security_role = 'N/A'
        user_group = 'N/A'
        project = 'N/A'

        # Record points that this privilege is directly assigned to the user
        if source.get('direct'):
            security_role = '--'
            user_group = '--'
            project = '--'
        else:
            # Record points that this privilege is inherited from a UserGroup
            if from_group := source.get('group'):
                security_role = '--'
                user_group = from_group.get('name', '--')
                project = '--'
            # Record points that this privilege is inherited from a SecurityRole
            if from_role := source.get('securityRole'):
                security_role = from_role.get('name', '--')
                user_group = '--'
                if on_project := source.get('project'):
                    project = on_project.get('name', '--')
        return user_group, security_role, project

    def _fetch(self) -> None:
        resp = license_api.get_license(
            connection=self._connection, node_name=self._node_name
        )
        retr = camel_to_snake(resp.json())
        self._init_variables(**retr)
        self.get_activation_info()

    def _read_contact_information_from_file(self, file_path: str) -> dict:
        """Read contact information from JSON or CSV files.

        Args:
            file_path (str): Path to the JSON or CSV file containing contact
                information.

        Returns:
            Dictionary with the read contact information.

        Raises:
            ValueError: If the file extension is not supported or if there's an
                error parsing the file.
        """

        file_ext = file_path.lower().split('.')[-1]

        try:
            if file_ext == 'json':
                with open(file_path) as f:
                    return json.load(f)
            if file_ext == 'csv':
                with open(file_path, newline='') as f:
                    reader = csv.DictReader(f)
                    data = next(reader)
                    for key, value in data.items():
                        if isinstance(value, str) and value.lower() in [
                            'true',
                            'false',
                        ]:
                            data[key] = value.lower() == 'true'
                    return data
            raise ValueError(
                f"Unsupported file extension: {file_ext}. Use .json or .csv files."
            )
        except Exception as e:
            raise ValueError(f"Error reading contact information from file: {str(e)}")

    def _prepare_body_for_information_update(
        self,
        operations_list: list,
        information: ContactInformation | dict | str,
        contact_key: str,
    ) -> None:
        if isinstance(information, ContactInformation):
            operations_list[0]['value']['contact'][contact_key] = information.to_dict()
        elif isinstance(information, dict):
            operations_list[0]['value']['contact'][contact_key] = information
        else:
            operations_list[0]['value']['contact'][contact_key] = (
                self._read_contact_information_from_file(information)
            )
        if 'confirmEmail' in operations_list[0]['value']['contact'][contact_key]:
            del operations_list[0]['value']['contact'][contact_key]['confirmEmail']

    def _check_node_name(self, node_name: str | None = None) -> None:
        """Check if the provided node name is valid and set the node name
        property.

        Args:
            node_name: Name of the node to check. If None, the first node
                will be used and assigned to node_name property.

        Raises:
            ValueError: If the node name is not found or if multiple nodes exist
                 but none specified.
        """
        env = Environment(connection=self._connection)
        nodes = env.nodes
        if node_name:
            if not any(node['name'] == node_name for node in nodes):
                raise ValueError(f"Node '{node_name}' not found in the environment.")
        else:
            if len(nodes) > 1:
                possible_nodes = [node['name'] for node in nodes]
                raise ValueError(
                    "Multiple nodes found in the environment. Please specify "
                    f"the node name. Possible nodes: {possible_nodes}"
                )
        self._node_name = node_name or nodes[0]['name']

    @property
    def activation_info(self) -> ActivationInfo:
        """ActivationInfo object with license activation details."""
        if not self._activation_info:
            self.get_activation_info()
        return self._activation_info

    @property
    def compliance_time(self) -> str:
        """Time of the last compliance check."""
        if not self._compliance_time:
            self.get_compliance_check()
        return self._compliance_time

    @property
    def compliance_type(self) -> int:
        """Type of the last compliance check."""
        if not self._compliance_type:
            self.get_compliance_check()
        return self._compliance_type

    @property
    def in_compliance(self) -> bool:
        """Whether the license is in compliance or not."""
        if self._in_compliance is None:
            self.get_compliance_check()
        return self._in_compliance

    @property
    def compliance_running(self) -> str:
        """Compliance check status of the license."""
        if not self._compliance_running:
            self.get_compliance_check_status()
        return self._compliance_running

    @property
    def audit_time(self) -> str:
        """Time of the last audit."""
        if not self._audit_time:
            self.get_audit()
        return self._audit_time



import logging

from mstrio import config
from mstrio.api import content_groups
from mstrio.connection import Connection
from mstrio.modeling.metric.metric_format import FormatProperty
from mstrio.server.project import Project
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.users_and_groups import User, UserGroup, UserOrGroup
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity
from mstrio.utils.helper import (
    delete_none_values,
    find_object_with_name,
    get_temp_connection,
    is_dashboard,
)
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.version_helper import class_version_handler, method_version_handler

logger = logging.getLogger(__name__)


@method_version_handler('11.3.1200')
def list_content_groups(
    connection: Connection,
    to_dictionary: bool = False,
    limit: int | None = None,
    name: str | None = None,
) -> list['ContentGroup'] | list[dict]:
    """Get a list of content groups.

    Args:
        connection (Connection): Strategy One connection object returned
            by 'connection.Connection()'
        to_dictionary (bool, optional): if True, return Content Groups as a
            list of dicts
        limit (int, optional): limit the number of elements returned
        name (str, optional): filter for content groups with names containing
            this value

    Returns:
        A list of content group objects or dictionaries representing them.
    """
    response = (
        content_groups.list_content_groups(connection=connection)
        .json()
        .get('contentGroups')
    )
    if name:
        response = [cg for cg in response if name in cg.get('name')]
    if limit:
        response = response[:limit]
    if to_dictionary:
        return response
    return [
        ContentGroup.from_dict(source=obj, connection=connection) for obj in response
    ]


@class_version_handler('11.3.1200')
class ContentGroup(Entity, CopyMixin, DeleteMixin):
    """Python representation of a Strategy One Content Group object"""

    _OBJECT_TYPE = ObjectTypes.CONTENT_BUNDLE
    _API_GETTERS = {
        **Entity._API_GETTERS,
        (
            'color',
            'opacity',
            'email_enabled',
            'recipients',
        ): content_groups.get_content_group,
    }
    _API_PATCH: dict = {
        (
            'comments',
            'owner',
        ): (objects_processors.update, 'partial_put'),
    }

    def __init__(
        self, connection: Connection, name: str | None = None, id: str | None = None
    ) -> None:
        """Initialize Content Group object by passing name or id.

        Args:
            connection (Connection): Strategy One connection object returned
                by `connection.Connection()`
            name (string, optional): name of Content Group
            id (string, optional): ID of Content Group
        """
        if id is None or id == '':
            if name is None:
                raise ValueError("Please specify either 'name' or 'id'.")
            content_group = find_object_with_name(
                connection=connection,
                cls=self.__class__,
                name=name,
                listing_function=list_content_groups,
            )
            id = content_group.get('id')
        super().__init__(
            connection=connection,
            object_id=id,
            name=name,
        )

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self._color = kwargs.get('color')
        self.opacity = kwargs.get('opacity')
        self.email_enabled = kwargs.get('email_enabled')
        self._recipients = kwargs.get('recipients')

    @classmethod
    def create(
        cls,
        connection: Connection,
        name: str,
        color: str,
        opacity: int = 100,
        email_enabled: bool = False,
        recipients: list[UserOrGroup] | None = None,
    ) -> 'ContentGroup':
        """Create a new content group.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`
            name (str): name of the content group
            color (str, optional): color of the content group, in hex format
                example '#ffe4e1' for misty rose pink
            opacity (int, optional): opacity percentage of the content group
                expressed in an int, ranges from 0-100, default is 100
            email_enabled (bool, optional): if True recipients will be notified
                about new content in this group via email, defaults to False
            recipients (list, optional): list of recipients of the content group
                represented as str containing ID or the User and UserGroup class
                objects

        Returns:
            ContentGroup object
        """
        recipients = [
            (
                {'id': recipient.id}
                if isinstance(recipient, (User, UserGroup))
                else {'id': recipient}
            )
            for recipient in (recipients or [])
        ]
        color = FormatProperty.Color(hex_value=color).server_value
        body = {
            'name': name,
            'color': color,
            'opacity': opacity,
            'emailEnabled': email_enabled,
            'recipients': recipients,
        }
        response = content_groups.create_content_group(
            connection=connection,
            body=body,
        ).json()
        if config.verbose:
            logger.info(
                f"Successfully created Content Group named: '{name}' with ID: '"
                f"{response.get('id')}'"
            )
        return cls.from_dict(source=response, connection=connection)

    def alter(
        self,
        name: str | None = None,
        comments: str | None = None,
        owner: str | User | None = None,
        color: str | None = None,
        opacity: int | None = None,
        email_enabled: bool | None = None,
        recipients: list[UserOrGroup] | None = None,
    ) -> None:
        """Alter the content group.

        Args:
            name (str, optional): name of the content group
            color (str, optional): color of the content group, in hex format
                example '#ffe4e1' for misty rose pink
            opacity (int, optional): opacity percentage of the content group
                expressed in an int, ranges from 0-100, default is 100
            email_enabled (bool, optional): if True recipients will be notified
                about new content in this group via email
            recipients (list, optional): list of recipients of the content group
                represented as str containing ID or the User and UserGroup class
                objects
            comments (str, optional): long description of the content group
            owner: (str, User, optional): owner of the content group
        """
        if isinstance(owner, User):
            owner = owner.id
        properties = {
            'comments': comments,
            'owner': owner,
        }
        properties = delete_none_values(properties, recursion=False)
        self._alter_properties(**properties)
        operations = [
            ('/name', name),
            (
                '/color',
                (
                    int(FormatProperty.Color(hex_value=color).server_value)
                    if color
                    else None
                ),
            ),
            ('/opacity', opacity),
            ('/emailEnabled', email_enabled),
        ]
        operation_list = []
        index = 1
        for path, value in operations:
            if value is not None:
                operation_list.append(
                    {'op': 'replace', 'path': path, 'value': value, 'id': index}
                )
                index += 1
        if recipients is not None:
            operation_list += self._prepare_recipients_for_alter(
                recipients=recipients, index_offset=len(operation_list) + 1
            )
        content_groups.update_content_group(
            connection=self.connection,
            id=self.id,
            body={'operationList': operation_list},
        )
        self.fetch()

    def _process_content(self, connection, content) -> Entity:
        """Process a single content item and return the corresponding object."""
        from mstrio.project_objects import Agent, Dashboard, Document, Report

        if content.get('type') == 3:
            return Report(connection=connection, id=content.get('id'))
        elif content.get('type') == 55:
            if content.get('subtype') == ObjectSubTypes.DOCUMENT_BOT.value:
                from mstrio.project_objects.bots import Bot

                return Bot(connection=connection, id=content.get('id'))
            elif content.get('subtype') in [
                ObjectSubTypes.DOCUMENT_AGENT.value,
                ObjectSubTypes.DOCUMENT_AGENT_UNIVERSAL.value,
            ]:
                return Agent(connection=connection, id=content.get('id'))
            elif is_dashboard(content.get('viewMedia')):
                return Dashboard(connection=connection, id=content.get('id'))
            else:
                return Document(connection=connection, id=content.get('id'))

    def get_contents(self, project_ids: list[str | Project]) -> list[Entity]:
        """Get contents of the content group.

        Args:
            project_ids (list): list of project IDs or Projects to search for
                contents

        Returns:
            A list of content objects.
        """

        project_ids = [
            proj.id if isinstance(proj, Project) else proj for proj in project_ids
        ]
        response = content_groups.get_content_group_contents(
            connection=self.connection,
            id=self.id,
            project_ids=project_ids,
        ).json()
        contents = []
        for project in response:
            project_change = None if project == self.connection.project_id else project
            temp_conn = get_temp_connection(self.connection, project_change)
            for content in response.get(project):
                contents.append(self._process_content(temp_conn, content))
        return contents

    def update_contents(
        self,
        content_to_add: list['Entity'] | None = None,
        content_to_remove: list['Entity'] | None = None,
    ) -> None:
        """Update contents of the content group.

        Args:
            content_to_add (list, optional): list of content objects to add to
                the content group. Provided as a list of Entity-based objects.
                Supported content types: Agent, Dashboard, Document, Report
            content_to_remove (list. optional): list of content objects to
                remove from the content group. Provided as a list of
                Entity-based objects.
        """
        if not content_to_add and not content_to_remove:
            raise ValueError(
                "Please provide at least one content object to add or remove."
            )
        content_to_add = content_to_add or []
        content_to_remove = content_to_remove or []
        operation_list = [
            {
                'op': 'add' if content in content_to_add else 'remove',
                'path': f'/{content.connection.project_id}',
                'value': [{'id': content.id, 'type': content._OBJECT_TYPE.value}],
                'id': index + 1,
            }
            for index, content in enumerate(content_to_add + content_to_remove)
        ]
        content_groups.update_content_group_contents(
            connection=self.connection,
            id=self.id,
            body={'operationList': operation_list},
        )

    def _prepare_recipients_for_alter(
        self, recipients: list[UserOrGroup], index_offset: int
    ) -> list[dict]:
        recipients = [
            (recipient.id if isinstance(recipient, (User, UserGroup)) else recipient)
            for recipient in recipients
        ]
        recipients_to_add = [
            r for r in recipients if r not in [r2.id for r2 in self.recipients]
        ]
        recipients_to_remove = [r.id for r in self.recipients if r.id not in recipients]
        operations = []
        for rec_id in recipients_to_add:
            operations.append(
                {
                    'op': 'add',
                    'path': '/recipients',
                    'value': [{'id': rec_id}],
                    'id': index_offset,
                }
            )
            index_offset += 1
        for rec_id in recipients_to_remove:
            operations.append(
                {
                    'op': 'remove',
                    'path': '/recipients',
                    'value': [{'id': rec_id}],
                    'id': index_offset,
                }
            )
            index_offset += 1

        return operations

    @property
    def color(self) -> str:
        """Color of the content group in hex format."""
        if self._color is None:
            self.fetch('color')
        if isinstance(self._color, int):
            self._color = FormatProperty.Color(server_value=str(self._color)).hex_value
        return self._color

    @property
    def recipients(self) -> list[UserOrGroup]:
        """Recipients of the content group."""
        if self._recipients is None:
            self.fetch('recipients')
        if self._recipients and isinstance(self._recipients[0], dict):
            self._recipients = [
                (
                    UserGroup.from_dict(source=recipient, connection=self.connection)
                    if recipient.get('group')
                    else User.from_dict(source=recipient, connection=self.connection)
                )
                for recipient in self._recipients
            ]
        return self._recipients



from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg="Error creating a language.")
def create_language(connection: Connection, body: dict, fields: str | None = None):
    """Create a new language.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        body (dict): JSON-formatted body of the new language
        fields(str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None

    Returns:
        HTTP response object returned by the Strategy One REST server."""
    return connection.post(
        endpoint='/api/languages',
        params={'fields': fields},
        json=body,
    )


@ErrorHandler(err_msg="Error getting languages list.")
def list_languages(
    connection: Connection,
    acl: list[str] = None,
    hidden: bool = False,
    is_language_supported: bool = None,
    fields: str | None = None,
    error_msg: str | None = None,
):
    """Get a list of languages.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        acl (list[str]): Access control list filter, defaults to None
        hidden (bool): Whether to include hidden objects, defaults to False
        is_language_supported: Whether to filter by languages derived from a
            list of supported languages, defaults to None, treated as True
            by REST API
        fields(str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None

    Returns:
        HTTP response object returned by the Strategy One REST server."""
    return connection.get(
        endpoint='/api/languages',
        params={
            'acl': acl,
            'hidden': str(hidden).lower(),
            'isLanguageSupported': (
                str(is_language_supported).lower()
                if is_language_supported
                else is_language_supported
            ),
            'fields': fields,
        },
    )


@ErrorHandler(err_msg="Error getting interface languages list.")
def list_interface_languages(connection: Connection, fields: str | None = None):
    """Get a list of interface languages.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        fields(str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model

    Returns:
        HTTP response object returned by the Strategy One REST server."""
    return connection.get(endpoint='/api/interfaceLanguages', params={'fields': fields})


@ErrorHandler(err_msg="Error getting language with ID: {id}")
def get_language(connection: Connection, id: str, fields: str | None = None):
    """Get a language by id.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (str): ID of the language to be retrieved from the server
        fields(str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None

    Returns:
        HTTP response object returned by the Strategy One REST server."""
    return connection.get(endpoint=f'/api/languages/{id}', params={'fields': fields})


@ErrorHandler(err_msg="Error getting formatting settings for language with ID: {id}")
def get_language_formatting_settings(
    connection: Connection, id: str, fields: str | None = None
):
    """Get formatting settings for a language by id.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (str): ID of the language of the settings to be retrieved from the
            server
        fields(str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None

    Returns:
        HTTP response object returned by the Strategy One REST server."""
    return connection.get(
        endpoint=f'/api/languages/{id}/formattingSettings',
        params={'fields': fields},
    )


@ErrorHandler(err_msg="Error updating formatting settings for language with ID: {id}")
def update_language_formatting_settings(
    connection: Connection, id: str, body: dict, fields: str | None = None
):
    """Update formatting settings for a language by id.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (str): ID of the language of the settings to be updated
        body (dict): JSON-formatted body of the settings update
        fields(str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None

    Returns:
        HTTP response object returned by the Strategy One REST server."""
    return connection.patch(
        endpoint=f'/api/languages/{id}/formattingSettings',
        params={'fields': fields},
        json=body,
    )


@ErrorHandler(err_msg="Error updating language with ID: {id}")
def update_language(
    connection: Connection, id: str, body: dict, fields: str | None = None
):
    """Update a language by id.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (str): ID of the language to be updated
        body (dict): JSON-formatted body of the language update
        fields(str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None

    Returns:
        HTTP response object returned by the Strategy One REST server."""
    return connection.patch(
        endpoint=f'/api/languages/{id}',
        params={'fields': fields},
        json=body,
    )



from requests import Response

from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg="Error getting license information.")
def get_license(
    connection: Connection, node_name: str, fields: str | None = None, error_msg=None
) -> Response:
    """Get the license information.

    Args:
        connection (Connection): Strategy One REST API connection object
        node_name (str): Name of the node for which to get license information
        fields (str | None, optional): Comma-separated list of fields to return
            in the response
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object. Expected status: 200.
    """
    endpoint = f'/api/license/iServer/nodes/{node_name}'
    params = {}

    if fields:
        params['fields'] = fields

    return connection.get(endpoint=endpoint, params=params)


@ErrorHandler(err_msg="Error getting license activation information.")
def get_license_activation_info(
    connection: Connection, node_name: str, fields: str | None = None, error_msg=None
) -> Response:
    """Get the license activation information.

    Args:
        connection (Connection): Strategy One REST API connection object
        node_name (str): Name of the node for which to get license activation
            information
        fields (str | None, optional): Comma-separated list of fields to return
            in the response
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object. Expected status: 200.
    """
    endpoint = f'/api/license/iServer/nodes/{node_name}/activation'
    params = {}

    if fields:
        params['fields'] = fields

    return connection.get(endpoint=endpoint, params=params)


@ErrorHandler(err_msg="Error updating license information.")
def update_license_information(
    connection: Connection, node_name: str, operations: list, error_msg=None
) -> Response:
    """Update the license information using operations list.

    Args:
        connection (Connection): Strategy One REST API connection object
        node_name (str): Name of the node for which to update license
            information
        operations (list): List of operations to perform on the license
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 204.
    """
    endpoint = f'/api/license/iServer/nodes/{node_name}'
    body = {'operationList': operations}

    return connection.patch(endpoint=endpoint, json=body)


@ErrorHandler(err_msg="Error updating license activation information.")
def update_license_activation(
    connection: Connection, node_name: str, operations: list, error_msg=None
) -> Response:
    """Update the license activation information.

    Args:
        connection (Connection): Strategy One REST API connection object
        node_name (str): Name of the node for which to update the license
            activation information
        operations (list): List of activation operations to be performed
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 204.
    """
    endpoint = f'/api/license/iServer/nodes/{node_name}/activation'
    body = {'operationList': operations}

    return connection.patch(endpoint=endpoint, json=body)


@ErrorHandler(err_msg="Error getting license activation XML file.")
def get_activation_xml_file(
    connection: Connection, node_name: str, fields: str | None = None, error_msg=None
) -> Response:
    """Get the activation XML file.

    Args:
        connection (Connection): Strategy One REST API connection object
        node_name (str): Name of the node for which to get activation XML file
        fields (str | None, optional): Comma-separated list of fields to return
            in the response
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object. Expected status: 200.
    """
    endpoint = f'/api/license/iServer/nodes/{node_name}/activation/xmlFile'
    params = {}

    if fields:
        params['fields'] = fields

    return connection.get(endpoint=endpoint, params=params)


@ErrorHandler(err_msg="Error getting license history.")
def get_license_history(
    connection: Connection, node_name: str, fields: str | None = None, error_msg=None
) -> Response:
    """Get the license history.

    Args:
        connection (Connection): Strategy One REST API connection object
        node_name (str): Name of the node for which to get license history
        fields (str | None, optional): Comma-separated list of fields to return
            in the response
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object. Expected status: 200.
    """
    endpoint = f'/api/license/iServer/nodes/{node_name}/history'
    params = {}

    if fields:
        params['fields'] = fields

    return connection.get(endpoint=endpoint, params=params)


@ErrorHandler(err_msg="Error getting license audit.")
def get_audit(
    connection: Connection, fields: str | None = None, error_msg=None
) -> Response:
    """Get the license audit.

    Args:
        connection (Connection): Strategy One REST API connection object
        fields (str | None, optional): Comma-separated list of fields to return
            in the response
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object. Expected status: 200.
    """
    endpoint = '/api/license/iServer/audit/result'
    params = {}

    if fields:
        params['fields'] = fields

    return connection.get(endpoint=endpoint, params=params)


@ErrorHandler(err_msg="Error getting compliance check results.")
def get_compliance_check(
    connection: Connection, fields: str | None = None, error_msg=None
) -> Response:
    """Get the license compliance check results.

    Args:
        connection (Connection): Strategy One REST API connection object
        fields (str | None, optional): Comma-separated list of fields to return
            in the response
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object. Expected status: 200.
    """
    endpoint = '/api/license/iServer/compliance/result'
    params = {}

    if fields:
        params['fields'] = fields

    return connection.get(endpoint=endpoint, params=params)


@ErrorHandler(err_msg="Error getting license compliance check status.")
def get_compliance_check_status(
    connection: Connection, fields: str | None = None, error_msg=None
) -> Response:
    """Get the license compliance check status.

    Args:
        connection (Connection): Strategy One REST API connection object
        fields (str | None, optional): Comma-separated list of fields to return
            in the response
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object. Expected status: 200.
    """
    endpoint = '/api/license/iServer/compliance/status'
    params = {}

    if fields:
        params['fields'] = fields

    return connection.get(endpoint=endpoint, params=params)


@ErrorHandler(err_msg="Error running license audit.")
def run_audit(connection: Connection, error_msg=None) -> Response:
    """Run the license audit.
    Args:
        connection (Connection): Strategy One REST API connection object
        node_name (str): Name of the node for which to run the license audit
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object. Expected status: 201.
    """
    endpoint = '/api/license/iServer/audit'
    return connection.post(endpoint=endpoint)


@ErrorHandler(err_msg="Error running license compliance check.")
def run_compliance_check(connection: Connection, error_msg=None) -> Response:
    """Run the license compliance check.

    Args:
        connection (Connection): Strategy One REST API connection object
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object. Expected status: 201.
    """
    endpoint = '/api/license/iServer/compliance'
    return connection.post(endpoint=endpoint)


@ErrorHandler(err_msg="Error getting license entitlements.")
def get_license_entitlements(
    connection: Connection,
    node_name: str,
    license_key: str,
    fields: str | None = None,
    error_msg=None,
) -> Response:
    """Get the license entitlements for a specific license key.

    Args:
        connection (Connection): Strategy One REST API connection object
        node_name (str): Name of the node for which to get license entitlements
        fields (str | None, optional): Comma-separated list of fields to return
            in the response
        license_key (str): License key for which to get entitlements
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object. Expected status: 200.
    """
    endpoint = f'/api/license/iServer/nodes/{node_name}/entitlements'
    params = {}
    headers = {'licenseKey': license_key}

    if fields:
        params['fields'] = fields

    return connection.get(endpoint=endpoint, headers=headers, params=params)


@ErrorHandler(err_msg="Error getting license privileges for user with ID:{id}.")
def get_privileges_for_user(
    connection: Connection,
    id: str,
    license_product: str,
    fields: str | None = None,
    error_msg=None,
) -> Response:
    """Get the license privileges for selected user.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (str): User ID for which to get license privileges
        license_product (str): License product for which to get privileges
        fields (str | None, optional): Comma-separated list of fields to return
            in the response
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object. Expected status: 200.
    """
    endpoint = f'/api/license/privileges/{id}'
    params = {'licenseProduct': license_product}

    if fields:
        params['fields'] = fields

    return connection.get(endpoint=endpoint, params=params)



from dataclasses import dataclass
from typing import TYPE_CHECKING

from mstrio.server.project import Project
from mstrio.utils.helper import Dictable

if TYPE_CHECKING:
    from mstrio.connection import Connection


@dataclass
class Node(Dictable):
    _FROM_DICT_MAP = {'projects': [Project.from_dict]}

    name: str | None = None
    address: str | None = None
    service_control: bool | None = None
    port: int | None = None
    status: str | None = None
    load: int | None = None
    projects: list[Project] | None = None
    default: bool | None = None

    @classmethod
    def from_dict(cls, source: dict, connection: 'Connection | None' = None):
        if not source.get('name'):
            source['name'] = source['node']
        if source.get('ipAddress'):
            source['address'] = source['ipAddress']
        return super().from_dict(source, connection=connection)


@dataclass
class Service(Dictable):
    id: str
    service: str
    port: int
    status: str
    tags: dict | None = None
    output: str | None = None


@dataclass
class ServiceWithNode(Dictable):
    node: str
    address: str
    id: str
    service_control: bool
    port: int
    status: str
    service_address: str
    tags: dict | None = None
    output: str | None = None



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from mstrio.object_management.translation import Translation
    from mstrio.utils.entity import Entity


class TranslationMixin:
    """TranslationMixin class adds translation managing and listing support
    for supported objects."""

    def add_translation(
        self: 'Entity',
        translations: list['Translation.OperationData'],
    ) -> list['Translation']:
        """Adds translations to the Object.

        Args:
            translations (list[OperationData]): list of translations to be added
                to the Object

        Returns:
            A list of translations for the Object.
        """
        from mstrio.object_management.translation import Translation

        return Translation.add_translation(
            connection=self.connection,
            id=self.id,
            object_type=self._OBJECT_TYPE.value,
            translations=translations,
            project_id=self.project_id if self.project_id else None,
        )

    def alter_translation(
        self: 'Entity',
        translations: list['Translation.OperationData'],
    ) -> None:
        """Alters translations of the Object.

        Args:
            translations (list[OperationData]): list of translations to be added
                to the Object
        """
        from mstrio.object_management.translation import Translation

        Translation.alter_translation(
            connection=self.connection,
            id=self.id,
            object_type=self._OBJECT_TYPE.value,
            translations=translations,
            project_id=self.project_id if self.project_id else None,
        )

    def remove_translation(
        self: 'Entity',
        translations: list['Translation.OperationData'],
    ) -> None:
        """Removes translations from the Object.

        Args:
            translations (list[OperationData]): list of translations to be added
                to the Object
        """
        from mstrio.object_management.translation import Translation

        Translation.remove_translation(
            connection=self.connection,
            id=self.id,
            object_type=self._OBJECT_TYPE.value,
            translations=translations,
            project_id=self.project_id if self.project_id else None,
        )

    def list_translations(
        self: 'Entity', languages: list | None = None, to_dictionary: bool = False
    ) -> list['Translation'] | list[dict]:
        """Lists translations for the Object.

        Args:
            languages (list, optional): list of languages to list the
                translations for, only translations from these languages
                will be listed.
                Languages in the list should be one of the following:
                    - lcid attribute of the language
                    - ID of the language
                    - Language class object
            to_dictionary (bool, optional): If True returns dict, by default
                (False) returns Translation objects

        Returns:
            A list of dictionaries representing translations for the Object or a
            list of Translation Objects."""
        from mstrio.object_management.translation import list_translations

        return list_translations(
            connection=self.connection,
            id=self.id,
            object_type=self._OBJECT_TYPE.value,
            project_id=self.project_id if self.project_id else None,
            languages=languages,
            to_dictionary=to_dictionary,
        )



from mstrio.connection import Connection
from mstrio.datasources.helpers import DBType, GatewayType
from mstrio.utils.entity import EntityBase
from mstrio.utils.enum_helper import get_enum_val
from mstrio.utils.helper import _prepare_objects as filter_objects
from mstrio.utils.response_processors import gateways
from mstrio.utils.version_helper import class_version_handler


def list_gateways(
    connection: Connection, to_dictionary: bool = False, **filters
) -> list['Gateway'] | list[dict]:
    """Get all gateways as list of Gateway objects or dictionaries.

    Optionally filter the gateways by specifying filters.

    Args:
        connection (Connection): Strategy One connection object
        to_dictionary (bool): If True returns a list of Gateway dicts,
           otherwise returns a list of Gateway objects
       **filters: Available filter parameters:
           ['id', 'name', 'gateway_type', 'db_type', 'is_certified']
    """
    return Gateway.list(connection, to_dictionary, **filters)


@class_version_handler('11.3.0960')
class Gateway(EntityBase):
    """Object representation of Strategy One Gateway

    Attributes:
        id: Gateway's ID
        name: Gateway's name
        gateway_type: Gateway's type
        db_type: Database type
        is_certified: Specifies if a gateway is certified
    """

    _FROM_DICT_MAP = {
        **EntityBase._FROM_DICT_MAP,
        'db_type': DBType,
        'gateway_type': GatewayType,
    }
    _API_GETTERS = {
        (
            'id',
            'name',
            'gateway_type',
            'db_type',
            'is_certified',
        ): gateways.get
    }

    def __init__(
        self,
        connection: Connection,
        id: str | None = None,
        name: str | None = None,
    ):
        """Initialize Driver object by passing ID or name.
        When `id` is provided, `name` is omitted.

        Args:
            connection (Connection): Strategy One connection object
            id (str): ID of Driver
            name (str): name of Driver
        """

        if id is None and name is None:
            raise ValueError(
                "Please specify either 'id' or 'name' parameter in the constructor."
            )

        if id is None:
            result: list[dict] = self.list(connection, name=name, to_dictionary=True)

            if result:
                object_data = result[0]
                self._init_variables(**object_data, connection=connection)
            else:
                raise ValueError(f"There is no Gateway named: '{name}'.")
        elif id == '':
            raise ValueError("Gateway's ID cannot be an empty string.")
        else:
            super().__init__(connection, id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)

        self.gateway_type = (
            GatewayType(kwargs.get('gateway_type'))
            if kwargs.get('gateway_type')
            else None
        )
        self.db_type = DBType(kwargs.get('db_type')) if kwargs.get('db_type') else None
        self.is_certified = kwargs.get('is_certified')

    @classmethod
    def list(
        cls, connection: Connection, to_dictionary: bool = False, **filters
    ) -> list[type['Gateway']] | list[dict]:
        """Get all gateways as list of Gateway objects or dictionaries.

        Optionally filter the gateways by specifying filters.

        Args:
            connection (Connection): Strategy One connection object
            to_dictionary (bool): If True returns a list of Gateway dicts,
               otherwise returns a list of Gateway objects
           **filters: Available filter parameters:
               ['id', 'name', 'gateway_type', 'db_type', 'is_certified']
        """

        if gateway_type := filters.get('gateway_type'):
            filters['gateway_type'] = get_enum_val(gateway_type, GatewayType)

        if db_type := filters.get('db_type'):
            filters['db_type'] = get_enum_val(db_type, DBType)

        objects = filter_objects(gateways.get_all(connection), filters)

        if to_dictionary:
            return objects

        return cls.bulk_from_dict(source_list=objects, connection=connection)



"""
Facilitate the migration of user data from one user to another such as
addresses or contacts.

1. Connect to the environment using data from the workstation
2. Get first user object based on provided name ('User Korean')
3. Get second user object based on provided name ('Tilda Austin')
4. Add addresses of the first user to the second user and remove them from
   the first user. Print each address before the operation
5. Get list of contacts for the first user
6. For each contact in the list alter the linked user to the second user
"""

from mstrio.connection import get_connection
from mstrio.users_and_groups.contact import list_contacts
from mstrio.users_and_groups.user import User

# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, project_name='MicroStrategy Tutorial')

# migrate addresses from one User to another
user_1 = User(conn, name='User Korean')
user_2 = User(conn, name='Tilda Austin')
for address in user_1.addresses:
    print(address)
    user_2.add_address(contact_address=address)
    user_1.remove_address(id=address['id'])

# migrate User contacts to another User
list_contacts = list_contacts(conn, linked_user=user_1)
for contact in list_contacts:
    contact.alter(linked_user=user_2)



import logging
from dataclasses import dataclass
from enum import Enum, auto
from typing import TYPE_CHECKING, Optional

from mstrio import config
from mstrio.api import metrics
from mstrio.connection import Connection
from mstrio.modeling.expression import Expression, ExpressionFormat
from mstrio.modeling.metric import Dimensionality, FormatProperty, MetricFormat
from mstrio.modeling.schema.helpers import (
    DataType,
    ObjectSubType,
    SchemaObjectReference,
)
from mstrio.object_management import Folder, SearchPattern, search_operations
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.users_and_groups.user import User
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity, MoveMixin
from mstrio.utils.enum_helper import AutoName, get_enum_val
from mstrio.utils.helper import (
    Dictable,
    delete_none_values,
    filter_params_for_func,
    find_object_with_name,
)
from mstrio.utils.resolvers import get_project_id_from_params_set
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.version_helper import method_version_handler
from mstrio.utils.vldb_mixin import ModelVldbMixin

if TYPE_CHECKING:
    from mstrio.modeling.metric import Metric
    from mstrio.server.project import Project

logger = logging.getLogger(__name__)


METRIC_SUBTYPES = (
    ObjectSubTypes.METRIC,
    ObjectSubTypes.METRIC_DMX,
    ObjectSubTypes.METRIC_TRAINING,
)


@method_version_handler(version='11.3.0500')
def list_metrics(
    connection: Connection,
    name: str | None = None,
    metric_type: ObjectTypes = ObjectTypes.METRIC,
    project: 'Project | str | None' = None,
    project_id: str | None = None,
    project_name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    search_pattern: SearchPattern | int = SearchPattern.CONTAINS,
    show_expression_as: ExpressionFormat | str = ExpressionFormat.TOKENS,
    **filters,
) -> list["Metric"] | list[dict]:
    """Get list of Metric objects or dicts with them.

    Optionally use `to_dictionary` to choose output format.

    Wildcards available for 'name':
        ? - any character
        * - 0 or more of any characters
        e.g. name_begins = ?onny will return Sonny and Tonny

    Args:
        connection: Strategy One connection object returned by
            `connection.Connection()`
        name (str, optional): characters that the metric name must
            begin with
        metric_type (ObjectTypes): one of metric subtypes: Metric or AggMetric
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        to_dictionary (bool, optional): If True returns dict, by default (False)
            returns Metric objects
        limit (integer, optional): limit the number of elements returned. If
            None all object are returned.
        search_pattern (SearchPattern enum or int, optional): pattern to
            search for, such as Begin With or Exactly. Possible values are
            available in ENUM `mstrio.object_management.SearchPattern`.
            Default value is CONTAINS (4).
        show_expression_as (ExpressionFormat, str): specify how expressions
            should be presented
            Available values:
                - `ExpressionFormat.TREE` or `tree`
                - `ExpressionFormat.TOKENS or `tokens` (default)
        **filters: Available filter parameters:
            id (str): Metric's ID
            name (str): Metric's name
            date_created (str): format: 2001-01-02T20:48:05.000+0000
            date_modified (str): format: 2001-01-02T20:48:05.000+0000
            version (str): Metric's version
            owner dict: e.g. {'id': <user's id>, 'name': <user's name>},
                with one or both of the keys: id, name
            acg (str | int): access control group

    Returns:
        list with Metric objects or list of dictionaries
    """
    proj_id = get_project_id_from_params_set(
        connection,
        project,
        project_id,
        project_name,
    )

    # For METRIC, exclude subtotal subtypes.
    # Did not set default value to allow explicitly passing METRIC
    if metric_type == ObjectTypes.METRIC:
        metric_type = list(METRIC_SUBTYPES)

    objects_ = search_operations.full_search(
        connection,
        object_types=metric_type,
        project=proj_id,
        name=name,
        pattern=search_pattern,
        limit=limit,
        **filters,
    )

    if to_dictionary:
        return objects_
    else:
        show_expression_as = (
            show_expression_as
            if isinstance(show_expression_as, ExpressionFormat)
            else ExpressionFormat(show_expression_as)
        )
        return [
            Metric.from_dict(
                source={**obj_, 'show_expression_as': show_expression_as},
                connection=connection,
                with_missing_value=True,
            )
            for obj_ in objects_
        ]


class DefaultSubtotals(Enum):
    AGGREGATION = SchemaObjectReference(
        sub_type=ObjectSubType.SYSTEM_SUBTOTAL,
        object_id='F225147A4CA0BB97368A5689D9675E73',
    )
    AVERAGE = SchemaObjectReference(
        sub_type=ObjectSubType.SYSTEM_SUBTOTAL,
        object_id='B328C60462634223B2387D4ADABEEB53',
    )
    COUNT = SchemaObjectReference(
        sub_type=ObjectSubType.SYSTEM_SUBTOTAL,
        object_id='078C50834B484EE29948FA9DD5300ADF',
    )
    GEOMETRIC_MEAN = SchemaObjectReference(
        sub_type=ObjectSubType.SYSTEM_SUBTOTAL,
        object_id='E1853D5A36C74F59A9F8DEFB3F9527A1',
    )
    MAXIMUM = SchemaObjectReference(
        sub_type=ObjectSubType.SYSTEM_SUBTOTAL,
        object_id='B1F4AA7DE683441BA559AA6453C5113E',
    )
    MEDIAN = SchemaObjectReference(
        sub_type=ObjectSubType.SYSTEM_SUBTOTAL,
        object_id='83A663067F7E43B2ABF67FD38ECDC7FE',
    )
    MINIMUM = SchemaObjectReference(
        sub_type=ObjectSubType.SYSTEM_SUBTOTAL,
        object_id='00B7BFFF967F42C4B71A4B53D90FB095',
    )
    MODE = SchemaObjectReference(
        sub_type=ObjectSubType.SYSTEM_SUBTOTAL,
        object_id='36226A4048A546139BE0AF5F24737BA8',
    )
    PRODUCT = SchemaObjectReference(
        sub_type=ObjectSubType.SYSTEM_SUBTOTAL,
        object_id='54E7BFD129514717A92BC44CF1FE5A32',
    )
    RESERVED = SchemaObjectReference(
        sub_type=ObjectSubType.SYSTEM_SUBTOTAL,
        object_id='F341109B11D5D528C00084916B98494F',
    )
    STANDARD_DEVIATION = SchemaObjectReference(
        sub_type=ObjectSubType.SYSTEM_SUBTOTAL,
        object_id='7FBA414995194BBAB2CF1BB599209824',
    )
    SUM_OF_WYA = SchemaObjectReference(
        sub_type=ObjectSubType.SYSTEM_SUBTOTAL,
        object_id='F7AE84A511D78008B00092BE4E571AD0',
    )
    TOTAL = SchemaObjectReference(
        sub_type=ObjectSubType.SYSTEM_SUBTOTAL,
        object_id='96C487AF4D12472A910C1ACACFB56EFB',
    )
    VARIANCE = SchemaObjectReference(
        sub_type=ObjectSubType.SYSTEM_SUBTOTAL,
        object_id='1769DBFCCF2D4392938E40418C6E065E',
    )
    WEIGHTED_YEARLY_AVERAGE = SchemaObjectReference(
        sub_type=ObjectSubType.SYSTEM_SUBTOTAL,
        object_id='F7AE852A11D78008B00092BE4E571AD0',
    )


@dataclass
class Threshold(Dictable):
    """Object that specifies a threshold

    Attributes:
        format (list[FormatProperty]): list of format properties for the
            threshold
        condition (Expression, optional): expression that specifies the
            condition on which the threshold is going to be applied
        name (string, optional): name of the threshold
        replace_text (string, optional): text that will replace the metric's
            value if the threshold is met
        semantics (Semantics, optional): value of an enumerator that specifies
            the type of replacement in the replace_text string
        scope (Scope, optional): value of an enumerator that specifies what the
            threshold applies to, the metric values, the subtotal value or both
        enable (bool): whether the threshold is enabled or not, True by default
    """

    class Scope(AutoName):
        METRIC_ONLY = auto()
        SUBTOTAL_ONLY = auto()
        METRIC_AND_SUBTOTAL = auto()

    class Semantics(AutoName):
        NUMBER = auto()
        TEXT = auto()
        PICTURE = auto()
        SYMBOL = auto()

    _FROM_DICT_MAP = {
        'format': [FormatProperty],
        'condition': Expression.from_dict,
        'semantics': Semantics,
        'scope': Scope,
    }

    format: list[FormatProperty]
    condition: Expression | None = None
    name: str | None = None
    replace_text: str | None = None
    semantics: Semantics | None = None
    scope: Scope | None = None
    enable: bool = True


class Metric(Entity, CopyMixin, MoveMixin, DeleteMixin, ModelVldbMixin):  # noqa: F811
    """Python representation of Strategy One Metric object.

    Attributes:
        id: metric's ID
        name: metric's name
        sub_type: string literal used to identify the type of a metadata object
        description: metric's description
        type: object type, ObjectTypes enum
        subtype: object subtype, ObjectSubTypes enum
        ext_type: object extended type, ExtendedType enum
        ancestors: list of ancestor folders
        date_created: creation time, DateTime object
        date_modified: last modification time, DateTime object
        destination_folder_id: a globally unique identifier used to distinguish
            between metadata objects within the same project
        is_embedded: If true indicates that the target object of this
            reference is embedded within this object. Alternatively if
            this object is itself embedded, then it means that the target
            object is embedded in the same container as this object.
        owner: User object that is the owner
        acg: access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: object access control list
        version: the version number this object is currently carrying
        expression: the Expression representing the metric's formula
        dimensionality: the object that specifies the dimensionality
        conditionality: the object that specifies the conditionality
        metric_subtotals: a list of the enabled subtotals for the metric
        aggregate_from_base: bool used to specify whether the metric aggregates
            from base
        formula_join_type: join type of the metric's formula, FormulaJoinType
            enumerator
        smart_total: whether the metric is decomposable or not, SmartTotal
            enumerator
        data_type: DataType object for metric values
        format: MetricFormat object that stores the formatting of the metric
        subtotal_from_base: bool used to specify whether the metric is
            a subtotal from base type
        column_name_alias: name for the column representing the metric in SQL
        metric_format_type: specifies whether the metric has HTML content,
            MetricFormatType enumerator
        thresholds: list of Threshold for the metric
        hidden: Specifies whether the object is hidden
    """

    @dataclass
    class Conditionality(Dictable):
        """Object that specifies the conditionality

        This class can only be used for simple metrics.
        For creation and altering of compound metrics, conditionality needs to
        be defined through the use of the Expression class.

        Attributes:
            filter (SchemaObjectReference, optional): reference to the filter
                to be used for the conditionality
            embed_method (EmbedMethod, optional): value of the enumerator that
                specifies the type of filter interaction between the metric and
                report filters
            prompt (SchemaObjectReference, optional): reference to the prompt
                to be used for the conditionality, only prompts of the type
                Prompt Object may be used
            remove_elements (bool): whether to remove related filter elements
        """

        class EmbedMethod(AutoName):
            BOTH_FILTERS_TOGETHER = auto()
            REPORT_INTO_METRIC_FILTER = auto()
            METRIC_INTO_REPORT_FILTER = auto()

        _FROM_DICT_MAP = {
            'filter': SchemaObjectReference.from_dict,
            'embedMethod': EmbedMethod,
            'prompt': SchemaObjectReference.from_dict,
        }

        filter: SchemaObjectReference | None = None
        embed_method: EmbedMethod | None = None
        prompt: SchemaObjectReference | None = None
        remove_elements: bool = False

    @dataclass
    class MetricSubtotal(Dictable):
        """Object that specifies the subtotal. User can use predefined subtotals
            using `DefaultSubtotals` enum, or specify custom made subtotals
            using SchemaObjectReference.

        Attributes:
            definition(SchemaObjectReference or DefaultSubtotals): Reference to
                the definition object
            implementation(SchemaObjectReference or DefaultSubtotals):
                Reference to the implementation object. Only used when the
                definition Subtotal is 'Total' to specify what the default
                subtotal type for the Total should be
        """

        _FROM_DICT_MAP = {
            'definition': SchemaObjectReference.from_dict,
            'implementation': SchemaObjectReference.from_dict,
        }

        def __init__(
            self,
            definition: SchemaObjectReference | DefaultSubtotals | None = None,
            implementation: SchemaObjectReference | DefaultSubtotals | None = None,
        ) -> None:
            self.definition = (
                definition.value
                if isinstance(definition, DefaultSubtotals)
                else definition
            )
            self.implementation = (
                implementation.value
                if isinstance(implementation, DefaultSubtotals)
                else implementation
            )

    class MetricFormatType(AutoName):
        RESERVED = auto()
        HTML_TAG = auto()
        LAST_ONE = auto()

    class SmartTotal(AutoName):
        DECOMPOSABLE_FALSE = auto()
        DECOMPOSABLE_TRUE = auto()

    class FormulaJoinType(AutoName):
        DEFAULT = auto()
        INNER = auto()
        OUTER = auto()

    _OBJECT_TYPE = ObjectTypes.METRIC
    _OBJECT_SUBTYPES = METRIC_SUBTYPES
    _API_GETTERS = {
        (
            'id',
            'sub_type',
            'name',
            'is_embedded',
            'description',
            'destination_folder_id',
            'expression',
            'dimty',
            'conditionality',
            'metric_subtotals',
            'aggregate_from_base',
            'formula_join_type',
            'smart_total',
            'data_type',
            'format',
            'subtotal_from_base',
            'column_name_alias',
            'metric_format_type',
            'thresholds',
        ): metrics.get_metric,
        (
            'abbreviation',
            'type',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl',
            'target_info',
            'hidden',
            'comments',
        ): objects_processors.get_info,
    }
    _API_PATCH = {
        (
            'id',
            'sub_type',
            'name',
            'is_embedded',
            'description',
            'destination_folder_id',
            'expression',
            'dimty',
            'conditionality',
            'metric_subtotals',
            'aggregate_from_base',
            'formula_join_type',
            'smart_total',
            'data_type',
            'format',
            'subtotal_from_base',
            'column_name_alias',
            'metric_format_type',
            'thresholds',
        ): (metrics.update_metric, 'partial_put'),
        (
            'folder_id',
            'hidden',
            'comments',
            'owner',
        ): (objects_processors.update, 'partial_put'),
    }
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'owner': User.from_dict,
        'sub_type': ObjectSubType,
        'expression': Expression.from_dict,
        'dimensionality': Dimensionality.from_dict,
        'conditionality': Conditionality.from_dict,
        'metric_subtotals': lambda source, connection: [
            Metric.MetricSubtotal.from_dict(content, connection) for content in source
        ],
        'formula_join_type': FormulaJoinType,
        'data_type': DataType.from_dict,
        'smart_total': SmartTotal,
        'format': MetricFormat.from_dict,
        'metric_format_type': MetricFormatType,
        'thresholds': lambda source, connection: [
            Threshold.from_dict(content, connection) for content in source
        ],
    }
    _REST_ATTR_MAP = {'dimty': 'dimensionality'}
    _MODEL_VLDB_API = {
        'GET_ADVANCED': metrics.get_vldb_settings,
        'PUT_ADVANCED': metrics.update_metric,
        'GET_APPLICABLE': method_version_handler('11.3.0900')(
            metrics.get_applicable_vldb_settings
        ),
    }

    @classmethod
    @method_version_handler('11.3.0500')
    def from_dict(
        cls,
        source: dict,
        connection: Optional['Connection'] = None,
        with_missing_value: bool = False,
    ):
        new_source = source.copy()
        new_source['dimensionality'] = source.get('dimty')
        return super().from_dict(
            source=new_source,
            connection=connection,
            with_missing_value=with_missing_value,
        )

    @method_version_handler('11.3.0500')
    def __init__(
        self,
        connection: Connection,
        id: str | None = None,
        name: str | None = None,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TOKENS,
    ) -> None:
        """Initializes a new instance of Metric class

        Args:
            connection (Connection): Strategy One connection object returned
                by `connection.Connection()`
            id (str, optional): Metric's ID. Defaults to None.
            name (str, optional): Metric's name. Defaults to None.
            show_expression_as (ExpressionFormat or str, optional):
                specify how expressions should be presented.
                Defaults to ExpressionFormat.TOKENS.

                Available values:
                - `ExpressionFormat.TREE` or `tree`
                - `ExpressionFormat.TOKENS or `tokens` (default)
        Note:
            Parameter `name` is not used when fetching. If only `name` parameter
            is provided, `id` will be found automatically if such object exists.

        Raises:
            ValueError: if both `id` and `name` are not provided
                or if Metric with the given `name` doesn't exist.
        """
        if id is None:
            if name is None:
                raise ValueError(
                    "Please specify either 'name' or 'id' parameter in the constructor."
                )

            metric = find_object_with_name(
                connection=connection,
                cls=self.__class__,
                name=name,
                listing_function=list_metrics,
                search_pattern=SearchPattern.EXACTLY,
            )
            id = metric['id']
        super().__init__(
            connection=connection,
            object_id=id,
            name=name,
            show_expression_as=show_expression_as,
        )

    @method_version_handler('11.3.0500')
    def _init_variables(self, default_value, **kwargs) -> None:
        super()._init_variables(default_value=default_value, **kwargs)
        self._sub_type = (
            ObjectSubType(kwargs.get('sub_type'))
            if kwargs.get('sub_type')
            else default_value
        )
        self._is_embedded = kwargs.get('is_embedded', default_value)
        self.destination_folder_id = kwargs.get('destination_folder_id', default_value)

        self.expression = (
            Expression.from_dict(exp)
            if (exp := kwargs.get('expression'))
            else default_value
        )
        self.dimensionality = (
            Dimensionality.from_dict(dimty)
            if (dimty := kwargs.get('dimensionality'))
            else default_value
        )
        self.conditionality = (
            Metric.Conditionality.from_dict(cond)
            if (cond := kwargs.get('conditionality'))
            else default_value
        )
        self.metric_subtotals = (
            [Metric.MetricSubtotal.from_dict(subtotal) for subtotal in subtotals]
            if (subtotals := kwargs.get('metric_subtotals'))
            else default_value
        )
        self.aggregate_from_base = kwargs.get('aggregate_from_base', default_value)
        self.formula_join_type = (
            Metric.FormulaJoinType(join_type)
            if (join_type := kwargs.get('formula_join_type'))
            else default_value
        )
        self.smart_total = (
            Metric.SmartTotal(tot)
            if (tot := kwargs.get('smart_total'))
            else default_value
        )
        self.data_type = (
            DataType.from_dict(dtype)
            if (dtype := kwargs.get('data_type'))
            else default_value
        )
        self.format = (
            MetricFormat.from_dict(form)
            if (form := kwargs.get('format'))
            else default_value
        )
        self.subtotal_from_base = kwargs.get('subtotal_from_base', default_value)
        self.column_name_alias = kwargs.get('column_name_alias', default_value)
        self.metric_format_type = (
            Metric.MetricFormatType(fromat_type)
            if (fromat_type := kwargs.get('metric_format_type'))
            else default_value
        )
        self.thresholds = (
            [Threshold.from_dict(threshold) for threshold in thresholds]
            if (thresholds := kwargs.get('thresholds'))
            else default_value
        )

        show_expression_as = kwargs.get('show_expression_as', 'tree')
        self.show_expression_as = (
            show_expression_as
            if isinstance(show_expression_as, ExpressionFormat)
            else ExpressionFormat(show_expression_as)
        )

    @classmethod
    @method_version_handler('11.3.0500')
    def create(
        cls,
        connection: 'Connection',
        name: str,
        sub_type: ObjectSubType | str,
        destination_folder: Folder | str,
        expression: Expression,
        description: str | None = None,
        is_embedded: bool = False,
        dimensionality: Dimensionality | None = None,
        conditionality: Conditionality | None = None,
        metric_subtotals: list[MetricSubtotal] | None = None,
        aggregate_from_base: bool = False,
        formula_join_type: FormulaJoinType | None = FormulaJoinType.DEFAULT,
        smart_total: SmartTotal | None = None,
        data_type: DataType | None = None,
        format: MetricFormat | None = None,
        subtotal_from_base: bool = False,
        column_name_alias: str | None = None,
        metric_format_type: MetricFormatType | None = None,
        thresholds: list[Threshold] | None = None,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TOKENS,
        hidden: bool | None = None,
    ) -> 'Metric':
        """Create a new metric with specified properties.

        Args:
            connection: Strategy One connection object returned
                by `connection.Connection()`
            name: metric's name
            sub_type: metric's sub_type
            destination_folder: A globally unique identifier used to
                distinguish between metadata objects within the same project.
                It is possible for two metadata objects in different projects
                to have the same Object Id
            description: metric's description
            is_embedded: If true indicates that the target object of this
                reference is embedded within this object. Alternatively if
                this object is itself embedded, then it means that the target
                object is embedded in the same container as this object.
            expression: the Expression representing the metric's formula
            dimensionality: the object that specifies the dimensionality
            conditionality: the object that specifies the conditionality
            metric_subtotals: a list of the enabled subtotals for the metric
            aggregate_from_base: bool used to specify whether the metric
                aggregates from base
            formula_join_type: join type of the metric's formula,
                FormulaJoinType enumerator
            smart_total: whether the metric is decomposable or not, SmartTotal
                enumerator
            data_type: DataType object for metric values
            format: MetricFormat object that stores the formatting of the metric
            subtotal_from_base: bool used to specify whether the metric is
                a subtotal from base type
            column_name_alias: name of the column representing the metric in SQL
            metric_format_type: specifies whether the metric has HTML content,
                MetricFormatType enumerator
            thresholds: list of Threshold for the metric
            show_expression_as (ExpressionFormat, str): specify how expressions
                should be presented
                Available values:
                - `ExpressionFormat.TREE` or `tree`
                - `ExpressionFormat.TOKENS or `tokens` (default)
            hidden (bool, optional): Specifies whether the object is hidden.
                Default value: False.

        Returns:
            Metric class object.
        """
        body = {
            'information': {
                'name': name,
                'subType': get_enum_val(sub_type, ObjectSubType),
                'isEmbedded': is_embedded,
                'description': description,
                'destinationFolderId': (
                    destination_folder.id
                    if isinstance(destination_folder, Folder)
                    else destination_folder
                ),
            },
            'expression': expression.to_dict() if expression else None,
            'dimty': dimensionality.to_dict() if dimensionality else None,
            'conditionality': conditionality.to_dict() if conditionality else None,
            'metricSubtotals': (
                [sub.to_dict() for sub in metric_subtotals]
                if metric_subtotals
                else None
            ),
            'aggregateFromBase': aggregate_from_base,
            'formulaJoinType': formula_join_type.value if formula_join_type else None,
            'smartTotal': smart_total.value if smart_total else None,
            'dataType': data_type.to_dict() if data_type else None,
            'format': format.to_dict() if format else None,
            'subtotalFromBase': subtotal_from_base,
            'columnNameAlias': column_name_alias,
            'metricFormatType': (
                metric_format_type.value if metric_format_type else None
            ),
            'thresholds': [x.to_dict() for x in thresholds] if thresholds else None,
        }
        body = delete_none_values(body, recursion=True)
        response = metrics.create_metric(
            connection,
            body=body,
            show_expression_as=get_enum_val(show_expression_as, ExpressionFormat),
        ).json()

        if config.verbose:
            logger.info(
                f"Successfully created metric named: '{name}' with ID:"
                f" '{response['id']}'"
            )

        metric = cls.from_dict(
            source={**response, 'show_expression_as': show_expression_as},
            connection=connection,
        )

        # Default value of 'hidden' attribute on IServer is False.
        # Because of that there is no reason to modify its value
        # after creation if 'hidden' parameter was provided as False.
        if hidden:
            metric.alter(hidden=hidden)

        return metric

    @method_version_handler('11.3.0500')
    def alter(
        self,
        name: str | None = None,
        destination_folder_id: str | None = None,
        expression: Expression | None = None,
        description: str | None = None,
        dimensionality: Dimensionality | None = None,
        conditionality: Conditionality | None = None,
        metric_subtotals: list[MetricSubtotal] | None = None,
        aggregate_from_base: bool = False,
        formula_join_type: FormulaJoinType | None = FormulaJoinType.DEFAULT,
        smart_total: SmartTotal | None = None,
        data_type: DataType | None = None,
        format: MetricFormat | None = None,
        subtotal_from_base: bool = False,
        column_name_alias: str | None = None,
        metric_format_type: MetricFormatType | None = None,
        thresholds: list[Threshold] | None = None,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TOKENS,
        hidden: bool | None = None,
        comments: str | None = None,
        owner: User | str | None = None,
    ):
        """Alter a metric's specified properties

        Args:
            name: metric's name
            destination_folder_id: A globally unique identifier used to
                distinguish between metadata objects within the same project.
                It is possible for two metadata objects in different projects
                to have the same Object Id.
            expression: the Expression representing the metric's formula
            description: metric's description
            dimensionality: the object that specifies the dimensionality
            conditionality: the object that specifies the conditionality
            metric_subtotals: a list of the enabled subtotals for the metric
            aggregate_from_base: bool used to specify whether the metric
                aggregates from base
            formula_join_type: join type of the metric's formula,
                FormulaJoinType enumerator
            smart_total: whether the metric is decomposable or not, SmartTotal
                enumerator
            data_type: DataType object for metric values
            format: MetricFormat object that stores the formatting of the metric
            subtotal_from_base: bool used to specify whether the metric is
                a subtotal from base type
            column_name_alias: name of the column representing the metric in SQL
            metric_format_type: specifies whether the metric has HTML content,
                MetricFormatType enumerator
            thresholds: list of Threshold for the metric
            show_expression_as (ExpressionFormat, str): specify how expressions
                should be presented
                Available values:
                - `ExpressionFormat.TREE` or `tree`
                - `ExpressionFormat.TOKENS or `tokens` (default)
            hidden: Specifies whether the metric is hidden
            owner: owner of the metric
        """
        name = name or self.name
        if isinstance(owner, User):
            owner = owner.id
        properties = filter_params_for_func(self.alter, locals(), exclude=['self'])
        self._alter_properties(**properties)

    @method_version_handler('11.3.0500')
    def to_dict(self, camel_case: bool = True) -> dict:
        result = super().to_dict(camel_case)
        result.pop('_showExpressionAs', None)
        if result.get('dimensionality'):
            result['dimty'] = result.pop('dimensionality')
        return result

    @property
    def sub_type(self):
        return self._sub_type

    @property
    def is_embedded(self):
        return self._is_embedded

    @property
    def project_id(self):
        return self._project_id if self._project_id else self.connection.project_id



from requests import Response

from mstrio.connection import Connection
from mstrio.utils.api_helpers import changeset_manager, unpack_information
from mstrio.utils.error_handlers import ErrorHandler


@unpack_information
@ErrorHandler(err_msg='Error getting prompt with ID: {id}.')
def get_prompt(
    connection: Connection,
    id: str,
    project_id: str | None = None,
    changeset_id: str | None = None,
    show_expression_as: list[str] | None = None,
    error_msg: str | None = None,
) -> Response:
    """Get a prompt by its ID.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (str): ID of the prompt to get
        project_id (str, optional): ID of the project to which the prompt
            belongs
        changeset_id (str, optional): ID of the changeset to use. Needed to
            return a prompt's definition within a specific changeset
        show_expression_as (list[str], optional): Specifies the format in
            which the expressions are returned in response
            Available values: 'tokens', 'tree'
            If omitted, the expression is returned in 'text' format
            If 'tree', the expression is returned in 'text' and 'tree' formats.
            If 'tokens', the expression is returned in 'text' and 'tokens'
            formats
        error_msg (str, optional): Custom error message

    Returns:
        HTTP response object. Returns 200 on success.
    """
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id
    return connection.get(
        endpoint=f'/api/model/prompts/{id}',
        params={'showExpressionAs': show_expression_as},
        headers={'X-MSTR-ProjectID': project_id, 'X-MSTR-MS-Changeset': changeset_id},
    )


@unpack_information
@ErrorHandler(err_msg='Error creating prompt with name: {name}.')
def create_prompt(
    connection: Connection,
    body: dict,
    show_expression_as: list[str] | None = None,
    error_msg: str | None = None,
) -> Response:
    """Create a new prompt.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        body (dict): JSON-formatted body of the new prompt
        show_expression_as (list[str], optional): Specifies the format in
            which the expressions are returned in response
            Available values: 'tokens', 'tree'
            If omitted, the expression is returned in 'text' format
            If 'tree', the expression is returned in 'text' and 'tree' formats.
            If 'tokens', the expression is returned in 'text' and 'tokens'
            formats.

    Returns:
        HTTP response object. Returns 201 on success.
    """
    with changeset_manager(connection) as changeset_id:
        return connection.post(
            endpoint='/api/model/prompts',
            headers={'X-MSTR-MS-Changeset': changeset_id},
            params={'showExpressionAs': show_expression_as},
            json=body,
        )


@unpack_information
@ErrorHandler(err_msg='Error updating prompt with ID: {id}.')
def update_prompt(
    connection: Connection,
    id: str,
    body: dict,
    show_expression_as: list[str] | None = None,
    error_msg: str | None = None,
) -> Response:
    """Update an existing prompt.

    Args:
        connection (Connection): Strategy One connection object
        id (str): ID of the prompt to update
        body (dict): JSON-formatted body with updated prompt data
        show_expression_as (list[str], optional): Specifies the format in
            which the expressions are returned in response
            Available values: 'tokens', 'tree'
            If omitted, the expression is returned in 'text' format
            If 'tree', the expression is returned in 'text' and 'tree' formats.
            If 'tokens', the expression is returned in 'text' and 'tokens'
            formats.

    Returns:
        HTTP response object. Returns 200 on success.
    """
    with changeset_manager(connection) as changeset_id:
        return connection.put(
            endpoint=f'/api/model/prompts/{id}',
            headers={
                'X-MSTR-MS-Changeset': changeset_id,
                'showExpressionAs': show_expression_as,
            },
            json=body,
        )


@ErrorHandler(err_msg='Error creating personal answer for prompt with ID: {id}.')
def create_personal_answer(
    connection: Connection,
    id: str,
    project_id: str,
    body: dict,
    instance_id: str | None = None,
    fields: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Create personal answers for a prompt.

    Args:
        connection (Connection): Strategy One connection object
        id (str): ID of the prompt to create a personal answer for
        project_id (str): ID of the project to which the prompt belongs
        body (dict): JSON-formatted body with personal answer data
        instance_id (str, optional): ID of the instance. Only needed for
            embedded prompts, not for standalone prompts
        fields (str, optional): Fields to include in the response
        error_msg (str, optional): Custom error message

    Returns:
        HTTP response object. Returns 201 on success.
    """
    return connection.post(
        endpoint=f'/api/prompts/{id}/personalAnswers',
        headers={
            'X-MSTR-ProjectID': project_id,
            'X-MSTR-InstanceID': instance_id,
            'fields': fields,
        },
        json=body,
    )


@ErrorHandler(err_msg='Error editing personal answer for prompt with ID: {id}.')
def edit_personal_answer(
    connection: Connection,
    id: str,
    project_id: str,
    personal_answer_id: str,
    body: dict,
    instance_id: str | None = None,
    fields: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Edit a personal answer for a prompt.

    Args:
        connection (Connection): Strategy One connection object
        id (str): ID of the prompt to edit a personal answer for
        project_id (str): ID of the project containing the prompt
        personal_answer_id (str): ID of the personal answer to edit
        body (dict): JSON-formatted body with updated personal answer data
        instance_id (str, optional): ID of the instance. Only needed for
            embedded prompts, not for standalone prompts
        fields (str, optional): Fields to include in the response
        error_msg (str, optional): Custom error message

    Returns:
        HTTP response object. Returns 200 on success."""
    return connection.patch(
        endpoint=f'/api/prompts/{id}/personalAnswers/{personal_answer_id}',
        headers={
            'X-MSTR-ProjectID': project_id,
            'X-MSTR-InstanceID': instance_id,
            'fields': fields,
        },
        json=body,
    )



import logging
from enum import Enum
from typing import TYPE_CHECKING

from pandas import DataFrame

from mstrio import config
from mstrio.api import administration as admin_api
from mstrio.api import monitors as monitors_api
from mstrio.helpers import IServerError
from mstrio.server.fence import Fence
from mstrio.server.fence import list_fences as _list_fences
from mstrio.server.project import Project, compare_project_settings
from mstrio.server.server import ServerSettings
from mstrio.server.storage import StorageService, StorageType
from mstrio.utils import helper
from mstrio.utils.resolvers import get_project_id_from_params_set
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.connection import Connection

logger = logging.getLogger(__name__)


class _LDAPBatchImport:
    """Class for handling LDAP batch import operations.

    This class is used to perform batch import operations on LDAP
    directories. It is not intended to be used directly by users.
    """

    class ImportStatus(Enum):
        NO_IMPORT = "No import"
        IN_PROGRESS = "In progress"
        FINISHED = "Finished"
        CANCELED = "Canceled"
        ERROR = "Error"
        UNKNOWN = "Unknown"  # local, not REST

    def __init__(self, parent: 'Environment'):
        """Initialize _LDAPBatchImport object.

        Args:
            parent: Environment object to which this import belongs.
        """
        self._parent = parent
        self._cached_status = self.ImportStatus.UNKNOWN
        self._status_data: dict | None = None

    def start(self) -> bool:
        """Start the LDAP batch import operation.

        Returns:
            bool: True if the import operation was started successfully,
                False otherwise.
        """
        res = admin_api.do_ldap_batch_import(self._connection)
        if not res.ok:
            self._cached_status = self.ImportStatus.ERROR
        else:
            self._cached_status = self.ImportStatus.IN_PROGRESS

        return res.ok

    def check_status(self) -> 'ImportStatus':
        """Check the status of the LDAP batch import operation.

        Returns:
            ImportStatus: The current status of the import operation.
        """
        res = admin_api.get_ldap_batch_import_status(self._connection)
        if not res.ok:
            self._cached_status = self.ImportStatus.UNKNOWN
        else:
            self._status_data = res.json()
            try:
                self._cached_status = self.ImportStatus(
                    stat := self._status_data.get('statusDescription')
                )
            except (ValueError, KeyError):
                logger.warning(
                    "Unexpected status of LDAP batch import received from server: "
                    f"{stat}."
                )
                self._cached_status = self.ImportStatus.UNKNOWN

        return self._cached_status

    def stop(self) -> bool:
        """Stop the LDAP batch import operation.

        Returns:
            bool: True if the import operation was stopped successfully,
        """
        res = admin_api.stop_ldap_batch_import(self._connection)
        if not res.ok:
            self._cached_status = self.ImportStatus.UNKNOWN
        elif self._cached_status != self.ImportStatus.FINISHED:
            self._cached_status = self.ImportStatus.CANCELED
        # if finished already, do not override

        return res.ok

    def get_status_data(self) -> dict:
        """Get the status data of the LDAP batch import operation.

        Returns:
            dict: The status data of the import operation.
        """
        self.check_status()
        return self._status_data

    @property
    def _connection(self) -> 'Connection':
        """Return the connection object taken from environment."""
        return self._parent.connection

    @property
    def status(self) -> 'ImportStatus':
        """Get the current status of the LDAP batch import operation.

        This value is cached. If you want to check the current server status,
        use `check_status()` method.

        Returns:
            ImportStatus: The current status of the import operation.
        """
        if not self._cached_status or self._cached_status == self.ImportStatus.UNKNOWN:
            self.check_status()

        return self._cached_status


@class_version_handler('11.3.0000')
class Environment:
    """Browse and manage Projects on the environment. List loaded
    projects, nodes (servers) and compare project settings on the
    environment. Browse and modify I-Server settings.

    Attributes:
        connection: A Strategy One connection object.
        server_settings: Intelligence Server settings object.
        nodes: List of I-Server nodes and their properties.
        node_names: List of I-Server node names.
    """

    def __init__(self, connection: 'Connection'):
        """Initialize Environment object.

        Args:
            connection: Strategy One connection object returned
                by `connection.Connection()`.
        """
        self.connection = connection
        self._nodes = None
        self._storage_service: StorageService | None = None
        self._ldap_batch_import_controller = _LDAPBatchImport(self)

    @property
    def server_settings(self) -> ServerSettings:
        """`ServerSettings` object storing I-Server settings.

        Settings can be listed by using `list_properties()` method.
        Settings can be modified directly by settings the values in the
        object.
        """
        if not hasattr(self, "_server_settings"):
            self._server_settings = ServerSettings(self.connection)
        return self._server_settings

    @server_settings.setter
    def server_settings(self, server_settings: ServerSettings) -> None:
        self._server_settings = server_settings

    def update_settings(self):
        """Update the current server settings saved in the ServerSettings
        object."""
        self.server_settings.update()

    def fetch_settings(self) -> None:
        """Fetch the current server settings from the environment."""
        self.server_settings.fetch()

    @property
    @method_version_handler('11.3.10')
    def storage_service(self) -> StorageService:
        """`StorageService` object with configuration of the Storage Service.
        object.
        """
        if not self._storage_service:
            response = admin_api.storage_service_get_configs(self.connection).json()
            self._storage_service = StorageService.from_dict(
                response['sharedFileStore'], self.connection
            )

        return self._storage_service

    def update_storage_service(
        self,
        storage_type: StorageType | None = None,
        alias: str | None = None,
        location: str | None = None,
        s3_region: str | None = None,
        aws_access_id: str | None = None,
        aws_secret_key: str | None = None,
        azure_storage_account_name: str | None = None,
        azure_secret_key: str | None = None,
        gcs_service_account_key: str | None = None,
        skip_validation: bool = False,
        validate_only: bool = False,
    ) -> None:
        """Update the storage service configuration on the environment.
        If new values aren't provided for any parameter, the config stored
        in the object will be used.

        Args:
            storage_type (StorageType, optional): Type of storage service.
            alias (str, optional): Alias of the storage configuration,
            location (str, optional): Storage location, e.g.
                bucket name for S3, absolute path of folder for File System
            s3_region (str, optional): S3 bucket region
            aws_access_id (str, optional): Access ID for AWS S3
            aws_secret_key (str, optional): Access key for AWS S3
            azure_storage_account_name (str, optional): Account name for Azure
            azure_secret_key (str, optional): Access key for Azure

            skip_validation: Whether to skip validation of the storage service
                configuration prior to updating.
            validate_only: If True, validate the configuration without updating.
        """
        validation_failed_msg = (
            "The storage service configuration failed validation. "
            "If this configuration is accepted by the I-Server, "
            "it will be applied, but the existing packages "
            "in the previous location will not be available."
        )
        if all(
            arg is None
            for arg in [
                storage_type,
                alias,
                location,
                s3_region,
                aws_access_id,
                aws_secret_key,
                azure_storage_account_name,
                azure_secret_key,
                gcs_service_account_key,
            ]
        ):
            config_delta = self.storage_service.to_dict()
        else:
            config_delta = {
                'type': storage_type.value if storage_type else None,
                'alias': alias,
                'location': location,
                's3Region': s3_region,
                'awsAccessId': aws_access_id,
                'awsSecretKey': aws_secret_key,
                'azureStorageAccountName': azure_storage_account_name,
                'azureSecretKey': azure_secret_key,
                'gcsServiceAccountKey': gcs_service_account_key,
            }
            config_delta = {k: v for k, v in config_delta.items() if v is not None}

        same_type = (
            'type' not in config_delta
            or config_delta['type'] == self.storage_service.type
        )
        old_config_dict = self._storage_service.to_dict() if same_type else {}
        new_config_dict = old_config_dict | config_delta
        storage_service_body = {'sharedFileStore': new_config_dict}
        try:
            admin_api.storage_service_validate_configs(
                self.connection, storage_service_body, error_msg=validation_failed_msg
            )
        except IServerError as e:
            self.fetch_storage_service()
            if e.http_code != 400 or not skip_validation:
                raise e

        if not validate_only:
            admin_api.storage_service_update_configs(
                self.connection, storage_service_body
            )
        self.fetch_storage_service()

    def fetch_storage_service(self) -> None:
        """Fetch the current configuration for Storage Service
        from the environment.
        """
        response = admin_api.storage_service_get_configs(self.connection).json()
        self._storage_service = StorageService.from_dict(
            response['sharedFileStore'], self.connection
        )

    def create_project(
        self, name: str, description: str | None = None, force: bool = False
    ) -> 'Project | None':
        """Create a new project on the environment.

        Args:
            name: Name of Project.
            description: Description of Application.
            force: If `True`, overrides the prompt.
        """
        return Project._create(self.connection, name, description, force)

    def list_projects(
        self, to_dictionary: bool = False, limit: int | None = None, **filters
    ) -> list["Project"] | list[dict]:
        """Return list of project objects or project dicts if
        `to_dictionary=True`. Optionally filter the Projects by specifying
        the `filters` keyword arguments.

        Args:
            to_dictionary: If True returns list of project dicts.
            limit: limit the number of elements returned. If `None`, all objects
                are returned.
            **filters: Available filter parameters: ['name', 'id',
                'description', 'date_created', 'date_modified', 'owner'].
        """
        return Project._list_projects(
            connection=self.connection,
            to_dictionary=to_dictionary,
            limit=limit,
            **filters,
        )

    def list_loaded_projects(
        self, to_dictionary: bool = False, **filters
    ) -> list["Project"] | list[dict]:
        """Return list of all loaded project objects or project dicts
        if `to_dictionary=True` that the user has access to. Optionally filter
        the Projects by specifying the `filters` keyword arguments.

        Args:
            to_dictionary: If True, returns list of project dicts
            **filters: Available filter parameters: ['acg', 'id', 'name',
                'status', 'alias', 'description', 'date_created',
                'date_modified', 'owner']
        """
        return Project._list_loaded_projects(
            connection=self.connection,
            to_dictionary=to_dictionary,
            **filters,
        )

    @method_version_handler('11.3.0800')
    def delete_server_object_cache(self) -> None:
        """Delete object cache for all projects on the environment."""
        for project in self.list_loaded_projects():
            project.delete_object_cache()
        if config.verbose:
            logger.info('Server object cache deleted.')

    @method_version_handler('11.3.0800')
    def delete_server_element_cache(self) -> None:
        """Delete element cache for all projects on the environment."""
        for project in self.list_loaded_projects():
            project.delete_element_cache()
        if config.verbose:
            logger.info('Server element cache deleted.')

    def list_nodes(
        self,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
        node_name: str | None = None,
    ) -> list[dict]:
        """Return a list of I-Server nodes and their properties. Optionally
        filter by `project` or `node_name`.

        Args:
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
            node_name (str, optional): Name of node
        """
        proj_id = get_project_id_from_params_set(
            self.connection,
            project,
            project_id,
            project_name,
            assert_id_exists=False,
        )
        response = monitors_api.get_node_info(
            self.connection, proj_id, node_name
        ).json()
        return response['nodes']

    def list_fences(
        self, to_dictionary: bool = False, limit: int | None = None, **filters
    ) -> list['Fence'] | list[dict]:
        """Get list of fences.

        Args:
            to_dictionary (bool, optional): If True returns dicts, by default
                (False) returns objects.
            limit (int, optional): limit the number of elements returned.
                If `None` (default), all objects are returned.
            **filters: Available filter parameters:
                ['name', 'id', 'type', 'rank']

        Returns:
            A list of content group objects or dictionaries representing them.
        """

        return _list_fences(self.connection, to_dictionary, limit, **filters)

    def is_loaded(
        self,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ) -> bool:
        """Check if project is loaded, by passing project ID or name,
        returns True or False.

        Args:
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name

        Returns:
            bool: True if project is loaded, False otherwise.
        """
        nodes = self.list_nodes(
            project=project, project_id=project_id, project_name=project_name
        )
        loaded = False

        for node in nodes:
            status = node['projects'][0]['status']
            loaded = status == 'loaded'
            if loaded:
                break

        return loaded

    def compare_settings(
        self,
        projects: list[str] | list["Project"] = None,
        show_diff_only: bool = False,
    ) -> DataFrame:
        """Compare project' settings to the first project in the
        provided list.

        Args:
            projects (list of names or project objects, optional): List
                of project objects or names to be compared. First element of
                list is the one to which the rest is compared. If None, all
                projects on the environment will be compared.
            show_diff_only(bool, optional): Whether to display all settings or
                only different from first project in list.

        Returns:
            Dataframe with values of selected project' settings.
        """

        def not_exist_warning(wrong_name):
            helper.exception_handler(
                f"Project '{wrong_name}' does not exist and will be skipped.",
                exception_type=Warning,
            )

        if projects:
            just_objects = [
                project for project in projects if isinstance(project, Project)
            ]
        else:
            just_objects = []
        if len(just_objects) == len(projects):
            return compare_project_settings(projects, show_diff_only)

        all_projects = self.list_projects()
        if isinstance(projects, list):
            if len(projects) < 2:
                helper.exception_handler(
                    "Provide more than one project object or name in list",
                    exception_type=TypeError,
                )

            # extract project names from either project object or strings
            project_names = [
                project.name if isinstance(project, Project) else str(project)
                for project in projects
            ]
            # filter only valid projects
            all_project_names = [project.name for project in all_projects]
            [
                not_exist_warning(a_name)
                for a_name in project_names
                if a_name not in all_project_names
            ]

            projects = list(
                filter(lambda project: project.name in project_names, all_projects)
            )
            projects = sorted(projects, key=lambda x: project_names.index(x.name))

        elif projects is None:
            projects = all_projects
        else:
            helper.exception_handler(
                "The 'projects' parameter needs to be a list of len > 1 or None.",
                exception_type=TypeError,
            )

        return compare_project_settings(projects, show_diff_only)

    def is_cluster(self):
        return len(self.nodes) > 1

    @property
    def nodes(self):
        if not self._nodes:
            self._nodes = self.list_nodes()

        return self._nodes

    @property
    def node_names(self):
        return [node['name'] for node in self.nodes]

    @property
    def ldap_batch_import(self):
        return self._ldap_batch_import_controller



"""This is the demo script to show how to export translations to database
and import them back. This script will not work without replacing parameters
with real values. Its basic goal is to present what can be done with this
module and to ease its usage.

1. Connect to the environment using data from workstation
2. Search for first ten metrics with "Revenue" in their name
3. Define export destination datasource (in this case "Tutorial Postgres")
4. Export translations to database without extra parameters
5. Import translations from database standard way
6. Export translations to database with specific languages only
7. Import translations from database with removing translations
   and auto matching target's ids
8. Export translations to database with extra columns
9. Import translations from database standard way
"""

from mstrio.connection import get_connection
from mstrio.datasources import DatasourceInstance
from mstrio.object_management import Translation
from mstrio.server import Language
from mstrio.object_management.search_operations import full_search
from mstrio.types import ObjectTypes

# Define a variable which can be later used in a script
PROJECT_NAME = 'MicroStrategy Tutorial'  # Project to connect to


# Create connection based on workstation data
CONN = get_connection(workstationData, project_name=PROJECT_NAME)

# Define variables which can be later used in a script
LANG_ID = '000004104F95EF3956E52781700C1E7A'
LANG_2_ID = '0000041D4F95EF3956E52781700C1E7A'
LANGUAGE_OBJECT = Language(connection=CONN, id=LANG_ID)
LANGUAGE_OBJECT_2 = Language(connection=CONN, id=LANG_2_ID)

# Search for first ten metrics with 'Revenue' in their name
search_result = full_search(
    connection=CONN,
    project=CONN.project_id,
    type=ObjectTypes.METRIC,
    name='Revenue',
    to_dictionary=False,
    limit=10,
)

# Export translations to database
# The flag force=True will be used in the examples
# That flag means no confirmation will be asked before overwriting the specified table
# If the flag is not used, the user will be asked for confirmation before overwriting

# First we need to define export destination datasource
# In this case it would be "Tutorial Postgres"
DATASOURCE = DatasourceInstance(connection=CONN, id='A23BBC514D336D5B4FCE919FE19661A3')

# Export translations to database without extra parameters
Translation.to_database_from_list(
    connection=CONN,
    object_list=search_result,
    table_name='test_table',
    datasource=DATASOURCE,
    force=True,
)

# Import translations from database standard way
Translation.add_translations_from_database(
    connection=CONN, table_name='test_table', datasource=DATASOURCE
)

# Export translations to database with specific languages only
Translation.to_database_from_list(
    connection=CONN,
    object_list=search_result,
    table_name='test_table',
    datasource=DATASOURCE,
    languages=[LANG_ID, LANG_2_ID],
    force=True,
)

# Import translations from database with removing translations and automatching target_ids
Translation.add_translations_from_database(
    connection=CONN,
    table_name='test_table',
    datasource=DATASOURCE,
    delete=True,
    automatch_target_ids=True,
)

# Export translations to database with extra columns
Translation.to_database_from_list(
    connection=CONN,
    object_list=search_result,
    table_name='test_table',
    datasource=DATASOURCE,
    add_object_creation_date=True,
    add_object_description=True,
    add_object_last_modified_date=True,
    add_object_path=True,
    add_object_version=True,
    force=True,
)

# Import translations from database standard way
Translation.add_translations_from_database(
    connection=CONN, table_name='test_table', datasource=DATASOURCE
)



"""This is the demo script meant to show how administrator can perform a
migration of objects from one environment to another.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from time import sleep

from mstrio.connection import Connection
from mstrio.object_management.migration import (
    Migration,
    list_migrations,
    list_migration_possible_content,
)
from mstrio.object_management import SearchObject, full_search
from mstrio.object_management.migration import (
    PackageType,
    MigrationPurpose,
    PackageSettings,
    PackageContentInfo,
    PackageConfig,
)
from mstrio.server.environment import StorageType
from mstrio.types import ObjectTypes
from mstrio.object_management.migration.package import (
    Action,
    ImportStatus,
    ProjectMergePackageTocView,
    ProjectMergePackageSettings,
    TranslationAction,
    PackageCertificationStatus,
    PackageStatus,
    ValidationStatus,
)
from mstrio.users_and_groups import User

# define connection to source environment
PROJECT_NAME = $project_name
ENV_URL_SOURCE = $env_url_source
U_NAME_SOURCE = $u_name_source
U_PASS_SOURCE = $u_pass_source
conn_source = Connection(
    base_url=ENV_URL_SOURCE,
    username=U_NAME_SOURCE,
    password=U_PASS_SOURCE,
    project_name=PROJECT_NAME,
)

# define connection to target environment
PROJECT_NAME = $project_name
ENV_URL_TARGET = $env_url_target
U_NAME_TARGET = $u_name_target
U_PASS_TARGET = $u_pass_target
conn_target = Connection(
    base_url=ENV_URL_TARGET,
    username=U_NAME_TARGET,
    password=U_PASS_TARGET,
    project_name=PROJECT_NAME,
)

# Configure Storage Service

# Configuration for Storage Service is stored in the storage_service attribute
# of the Environment object.
# This step should be done for source and target environments
env = conn_source.environment
print(env.storage_service)

# Make changes to the storage service configuration
env.storage_service.aws_access_id = "new_access_id"
env.storage_service.aws_secret_key = "new_secret_key"
env.storage_service.location = "new_s3_location"
env.storage_service.s3_region = "new_s3_region"
# Apply changes
env.update_storage_service()


# For target environment
env_target = conn_target.environment

# Make changes to the storage service configuration
env_target.storage_service.aws_access_id = "new_access_id"
env_target.storage_service.aws_secret_key = "new_secret_key"
env_target.storage_service.location = "new_s3_location"
env_target.storage_service.s3_region = "new_s3_region"
# Apply changes
env_target.update_storage_service()

# Alternatively, update the storage service configuration with a single call
STORAGE_ALIAS = $storage_alias
STORAGE_LOCATION = $storage_location
env_target.update_storage_service(
    storage_type=StorageType.FILE_SYSTEM, location=STORAGE_LOCATION, alias=STORAGE_ALIAS
)

# list_migrations
all_migs = list_migrations(conn_source)
first_5_migs = list_migrations(conn_source, limit=5)
only_file_migrations = list_migrations(
    conn_source, migration_purpose=MigrationPurpose.FROM_FILE
)
migrations_list_dict = list_migrations(conn_source, to_dictionary=True)

print(all_migs)

# Create an Object migration

# The PackageConfig object stores information on which objects should be
# migrated and how to resolve conflicts.
# The options are of type Enum with all possible values listed.
package_settings = PackageSettings(
    Action.USE_EXISTING,
    PackageSettings.UpdateSchema.RECAL_TABLE_LOGICAL_SIZE,
    PackageSettings.AclOnReplacingObjects.REPLACE,
    PackageSettings.AclOnNewObjects.KEEP_ACL_AS_SOURCE_OBJECT,
)

# Check what object types are allowed to object migration
migrated_types = list_migration_possible_content(conn_source, PackageType.OBJECT)
print(migrated_types)

# Now try to list every elements that have word "Day" in name and can be
# migrated with object migration
search_results = full_search(
    conn_source, conn_source.project_id, object_types=migrated_types, name="Day"
)
print(search_results)
# We can select from there one element and add to new migration body
package_content_from_search = PackageContentInfo(
    id=search_results[0]['id'],
    type=search_results[0]['type'],
    action=Action.USE_EXISTING,
    include_dependents=False,
)

# Include objects by their ID and type
MIGRATED_METRIC_ID = $migrated_metric_id
package_content_from_object = PackageContentInfo(
    id=MIGRATED_METRIC_ID,
    type=ObjectTypes.METRIC,
    action=Action.USE_EXISTING,
    include_dependents=True,
)

package_config = PackageConfig(
    package_settings, [package_content_from_search, package_content_from_object]
)

print(package_config.to_dict())

# Also we can build PackageConfig from search results
# Specifying object_action_map and object_dependents_map
object_action_map=[(ObjectTypes.METRIC,Action.USE_OLDER), (ObjectTypes.ATTRIBUTE, Action.USE_NEWER)]
object_dependents_map=[(ObjectTypes.METRIC, True), (ObjectTypes.ATTRIBUTE, True)]
package_config_search_results = Migration.build_package_config(
    conn_source,
    search_results,
    package_settings,
    object_action_map,
    object_dependents_map)
print(package_config_search_results.to_dict())

# Another way to build PackageConfig is to use SearchObject
SEARCH_OBJ_ID = $search_obj_id
search_obj = SearchObject(connection=conn_source, id=SEARCH_OBJ_ID)

package_config_search_obj = Migration.build_package_config(
    conn_source,
    content=search_obj,
    package_settings=package_settings)
print(package_config_search_obj.to_dict())

# Create the migration
my_obj_mig = Migration.create_object_migration(
    connection=conn_source,
    toc_view=package_config,
    name="object_mig",
    project_id=conn_source.project_id,
)

# Wait for the migration to be created
while my_obj_mig.package_info.status == PackageStatus.CREATING:
    sleep(2)
    my_obj_mig.fetch()

# Check migration status
print(my_obj_mig.package_info.status)

my_obj_mig.alter_migration_info(name=f"{my_obj_mig.name}_altered")

# Certify the migration
my_obj_mig.certify(
    status=PackageCertificationStatus.CERTIFIED,
    creator=User(conn_source, 'Administrator'),
)

print(my_obj_mig.package_info.certification)

TARGET_PROJECT_NAME = $target_project_name
my_obj_mig.trigger_validation(
    target_env=conn_target, target_project_name=TARGET_PROJECT_NAME
)

# Wait for the validation to be completed
while my_obj_mig.validation.status == ValidationStatus.VALIDATING:
    sleep(2)
    my_obj_mig.fetch()

# Check migration validation status
print(my_obj_mig.validation)

# Migrate to target environment
my_obj_mig.migrate(target_env=conn_target, target_project_name=TARGET_PROJECT_NAME)

# Get migration object on target environment
my_obj_mig_target = Migration(conn_target, id=my_obj_mig.id)

# Waiting for migration to finish
while my_obj_mig_target.import_info.status in [ImportStatus.IMPORTING, ImportStatus.PENDING]:
    sleep(2)
    my_obj_mig_target.fetch()

# Print final ImportStatus here
print(my_obj_mig_target.import_info.status)

REUSE_TARGET_PROJECT_NAME = $reuse_target_project_name
my_obj_mig.reuse(target_env=conn_target, target_project_name=REUSE_TARGET_PROJECT_NAME)

REVERSE_TARGET_PROJECT_ID = $reverse_target_project_id
my_obj_mig.reverse(target_env=conn_target, target_project_id=REVERSE_TARGET_PROJECT_ID)

# Download migration package to local storage
downloaded = my_obj_mig.download_package(save_path="./")

# Import a migration package in target environment without use of S3 bucket
# or other shared folder
SOURCE_FILE_PATH = downloaded["filepath"]
FILE_MIG_TARGET_PROJECT_NAME = $file_mig_target_project_name
from_file_mig = Migration.migrate_from_file(
    connection=conn_target,
    file_path=SOURCE_FILE_PATH,
    package_type=PackageType.OBJECT,
    name="object_mig_from_local_storage",
    target_project_name=FILE_MIG_TARGET_PROJECT_NAME,
)

# Create an Administration migration

# Create PackageConfig with information what object should be migrated and how.
# The options are of type Enum with all possible values listed.
package_settings_adm = PackageSettings(
    default_action=Action.USE_EXISTING,
    acl_on_replacing_objects=PackageSettings.AclOnReplacingObjects.REPLACE,
    acl_on_new_objects=PackageSettings.AclOnNewObjects.KEEP_ACL_AS_SOURCE_OBJECT,
)
MIGRATED_USER_ID = $migrated_user_id
package_content_info_adm = PackageContentInfo(
    id=MIGRATED_USER_ID,
    type=ObjectTypes.USER,
    action=Action.USE_EXISTING,
    include_dependents=True,
)

package_config_adm = PackageConfig(package_settings_adm, package_content_info_adm)

my_adm_mig = Migration.create_admin_migration(
    connection=conn_source, toc_view=package_config_adm, name="administration_mig"
)

# Wait for the migration to be created
while my_adm_mig.package_info.status == PackageStatus.CREATING:
    sleep(2)
    my_adm_mig.fetch()

# Check migration status
print(my_adm_mig.package_info.status)

my_adm_mig.trigger_validation(target_env=conn_target)

# Wait for the validation to be completed
while my_obj_mig.validation.status == ValidationStatus.VALIDATING:
    sleep(2)
    my_adm_mig.fetch()

# Check migration validation status
print(my_adm_mig.validation)

# Create a Project Merge migration

# Define project merge settings
proj_merge_settings = ProjectMergePackageSettings(
    translation_action=TranslationAction.NOT_MERGED,
    acl_on_replacing_objects=PackageSettings.AclOnReplacingObjects.USE_EXISTING,
    acl_on_new_objects=PackageSettings.AclOnNewObjects.KEEP_ACL_AS_SOURCE_OBJECT,
    validate_dependencies=False,
    default_action=Action.USE_EXISTING,
    object_type_actions=[
        {"type": 3, "subtype": 776, "action": "replace"},
        {"type": 3, "subtype": 779, "action": "replace"},
        {"type": 55, "action": "keep_both"},
        {"type": 55, "action": "keep_both"},
    ],
)

proj_merge_toc_view = ProjectMergePackageTocView(proj_merge_settings)

print(proj_merge_toc_view.to_dict())

MERGE_TARGET_PROJECT_NAME = $merge_target_project_name
my_new_proj_merge_mig = Migration.create_project_merge_migration(
    connection=conn_source,
    toc_view=proj_merge_toc_view,
    name="merge_mig",
    project_name=MERGE_TARGET_PROJECT_NAME,
)

# Delete all migrations on source environment
migs = list_migrations(conn_source)
for mig in migs:
    mig.delete(force=True)

all_migs = list_migrations(conn_source)
print(all_migs)



# flake8: noqa
from .cluster import Cluster, GroupBy, ServiceAction
from .environment import Environment
from .language import Language, list_interface_languages, list_languages
from .license import (
    ActivationInfo,
    ContactInformation,
    InstallationUse,
    License,
    MachineInfo,
    PrivilegeInfo,
    Product,
    Record,
    UserLicense,
)
from .node import Node
from .project import (
    DuplicationConfig,
    IdleMode,
    Project,
    ProjectDuplication,
    ProjectDuplicationStatus,
    ProjectInfo,
    ProjectSettings,
    ProjectStatus,
    compare_project_settings,
    list_projects,
    list_projects_duplications,
)
from .server import ServerSettings

# isort: off
from .job_monitor import (
    Job,
    JobStatus,
    JobType,
    kill_all_jobs,
    kill_jobs,
    list_jobs,
    ObjectType,
    PUName,
    SubscriptionType,
)

# isort: on



"""This is the demo script to show how administrator can delete unused managed
objects in a specific project.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""
from mstrio.connection import get_connection
from mstrio.server import Project

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert name of project to interact with

# get project with a given name and a connection
conn = get_connection(workstationData)
project = Project(connection=conn, name=PROJECT_NAME)
conn.select_project(project=project)

# you can delete unused managed objects in a specific project by simply calling
# the method below on a Project instance
#
# [DISCLAIMER]
#
# This is a very long and resource intensive operation and should be run with
# caution and only if necessary.
project.delete_unused_managed_objects()

# By default only objects with no dependents are deleted. You can change this
# behavior by enabling `try_force_delete` flag on a method. Then if the method
# was not able to deduce whether a managed object is unused, it will try to
# remove it.
project.delete_unused_managed_objects(try_force_delete=True)

# script logs will by default show how many managed objects were found and
# how many were deleted. If any were not deleted, it will log their IDs.

# you can also integrate the result of this operation in your script's logic by
# providing a flag to return all failed items instead of plain `boolean` value
FAILED_ITEMS = project.delete_unused_managed_objects(return_failed_items=True)

# The method returns boolean representing whether the operation removed all
# unused managed objects found by default. With a flag mentioned above set to
# True, it will return a list of dicts containing information about the
# objects that were not removed.
#
# You can gather them and try the below snippet to see what might have happened
from mstrio.object_management.object import Object

FAILED_ITEMS = project.delete_unused_managed_objects(return_failed_items=True)

for item in FAILED_ITEMS:
    try:
        obj = Object.from_dict(item, connection=conn)
        has_deps = obj.has_dependents()
        print(has_deps)  # will print `True/False` or will fail and be caught
        if not has_deps:
            obj.delete(force=True)  # will delete the object or fail and be caught
    except Exception as err:
        print(err)



from dataclasses import dataclass
from enum import Enum
from logging import getLogger

from mstrio.utils.helper import Dictable

logger = getLogger(__name__)


class StorageType(Enum):
    """Enumeration of Storage Service types."""

    UNSET = "unset"
    UNKNOWN = "unknown"
    FILE_SYSTEM = "file_system"
    S3 = "S3"
    AZURE = "Azure"
    GCS = "GCS"


@dataclass
class StorageService(Dictable):
    """StorageService configuration of environment.
    Attributes:
        type (StorageType): Type of storage (e.g. file system, S3)
        alias (str, optional): Alias of the storage configuration,
        location (str, optional): Storage location, e.g.
            bucket name for S3, absolute path of folder for File System
        s3_region (str, optional): S3 bucket region
        aws_access_id (str, optional): Access ID for AWS S3
        aws_secret_key (str, optional): Access key for AWS S3
        azure_storage_account_name (str, optional): Account name for Azure
        azure_secret_key (str, optional): Access key for Azure
        configured (bool): whether Storage Service is configured
    """

    _FROM_DICT_MAP = {
        'type': StorageType,
    }

    type: StorageType = StorageType.UNSET
    alias: str | None = None
    location: str | None = None
    s3_region: str | None = None
    aws_access_id: str | None = None
    aws_secret_key: str | None = None
    azure_storage_account_name: str | None = None
    azure_secret_key: str | None = None
    gcs_service_account_key: str | None = None
    configured: bool = False

    # Give a warning when changing type
    # This is equivalent to the warning present in Workstation
    def __setattr__(self, name, value):
        if (
            name == 'type'
            and hasattr(self, 'type')
            and value != self.type
            and self.type != StorageType.UNSET
        ):
            logger.warning(
                "Changing the storage service type. The existing packages "
                "in the previous location will not be available."
            )
        super().__setattr__(name, value)



from typing import Any, Callable, TypeVar

T = TypeVar('T')


def remove_duplicate_objects(objects: list[T], getter: Callable[[T], Any]) -> list[T]:
    """
    Remove duplicate objects from a list of objects
    based on the value returned by the getter function.
    This function assumes that getter will always return
    a comparable value and will not fail.

    Args:
        objects (list[T]): List of objects to filter.
        getter (Callable[[T], Any]): Function to get the value from the object
            by which the objects should be compared.

    Returns:
        list[T]: List of unique objects.
    """
    seen = set()
    unique_objects = []

    for obj in objects:
        value = getter(obj)
        if value not in seen:
            seen.add(value)
            unique_objects.append(obj)

    return unique_objects
